"""
The ``@docker`` connector allows you to build Docker images, or modify running
Docker containers, using ``pyinfra``. You can pass either an image name or
existing container ID:

+ Image - will create a container from the image, execute operations and save
    into a new image
+ Existing container ID - will simply execute operations against the container,
    leaving it up afterwards


.. code:: shell

    # A Docker base image must be provided
    pyinfra @docker/alpine:3.8 ...

    # pyinfra can run on multiple Docker images in parallel
    pyinfra @docker/alpine:3.8,@docker/ubuntu:bionic ...

    # Execute against a running container
    pyinfra @docker/2beb8c15a1b1 ...
"""

import json
import os
from tempfile import mkstemp
from typing import TYPE_CHECKING

import click

from pyinfra import local, logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.connectors import BaseConnectorMeta
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io
from pyinfra.progress import progress_spinner

from .local import run_shell_command as run_local_shell_command
from .util import make_unix_command_for_host

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class Meta(BaseConnectorMeta):
    handles_execution: bool = True
    keys_prefix: str = "docker"

    class DataKeys:
        identifier = "ID of container or image to target"
        container_id = "ID of container to target, overrides ``docker_identifier``"


DATA_KEYS = Meta.keys()


def make_names_data(identifier=None):
    if not identifier:
        raise InventoryError("No docker base ID provided!")

    yield (
        "@docker/{0}".format(identifier),
        {DATA_KEYS.identifier: identifier},
        ["@docker"],
    )


def _find_start_docker_container(container_id):
    docker_info = local.shell("docker container inspect {0}".format(container_id))
    docker_info = json.loads(docker_info)[0]
    if docker_info["State"]["Running"] is False:
        logger.info("Starting stopped container: {0}".format(container_id))
        local.shell("docker container start {0}".format(container_id))


def _start_docker_image(image_name):
    try:
        return local.shell(
            "docker run -d {0} tail -f /dev/null".format(image_name),
            splitlines=True,
        )[
            -1
        ]  # last line is the container ID
    except PyinfraError as e:
        raise ConnectError(e.args[0])


def connect(state: "State", host: "Host"):
    docker_container_id = host.data.get(DATA_KEYS.container_id)
    if docker_container_id:  # user can provide a docker_container_id
        host.connector_data["docker_container_no_disconnect"] = True
        host.connector_data["docker_container_id"] = docker_container_id
        return True

    docker_identifier = getattr(host.data, DATA_KEYS.identifier)
    with progress_spinner({"prepare docker container"}):
        try:
            # Check if the provided @docker/X is an existing container ID
            _find_start_docker_container(docker_identifier)
        except PyinfraError:
            container_id = _start_docker_image(docker_identifier)
        else:
            container_id = docker_identifier
            host.connector_data["docker_container_no_disconnect"] = True

    host.connector_data["docker_container_id"] = container_id
    return True


def disconnect(state, host):
    container_id = host.connector_data["docker_container_id"]

    if host.connector_data.get("docker_container_no_disconnect"):
        logger.info(
            "{0}docker build complete, container left running: {1}".format(
                host.print_prefix,
                click.style(container_id, bold=True),
            ),
        )
        return

    with progress_spinner({"docker commit"}):
        image_id = local.shell("docker commit {0}".format(container_id), splitlines=True)[-1][
            7:19
        ]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

    with progress_spinner({"docker rm"}):
        local.shell(
            "docker rm -f {0}".format(container_id),
        )

    logger.info(
        "{0}docker build complete, image ID: {1}".format(
            host.print_prefix,
            click.style(image_id, bold=True),
        ),
    )


def run_shell_command(
    state: "State",
    host: "Host",
    command,
    get_pty=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    **command_kwargs,
):
    container_id = host.connector_data["docker_container_id"]

    command = make_unix_command_for_host(state, host, command, **command_kwargs)
    command = QuoteString(command)

    docker_flags = "-it" if get_pty else "-i"
    docker_command = StringCommand(
        "docker",
        "exec",
        docker_flags,
        container_id,
        "sh",
        "-c",
        command,
    )

    return run_local_shell_command(
        state,
        host,
        docker_command,
        timeout=timeout,
        stdin=stdin,
        success_exit_codes=success_exit_codes,
        print_output=print_output,
        print_input=print_input,
        return_combined_output=return_combined_output,
    )


def put_file(
    state: "State",
    host: "Host",
    filename_or_io,
    remote_filename,
    remote_temp_filename=None,  # ignored
    print_output=False,
    print_input=False,
    **kwargs,  # ignored (sudo/etc)
):
    """
    Upload a file/IO object to the target Docker container by copying it to a
    temporary location and then uploading it into the container using ``docker cp``.
    """

    fd, temp_filename = mkstemp()

    try:
        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(temp_filename, "wb") as temp_f:
                data = file_io.read()

                if isinstance(data, str):
                    data = data.encode()

                temp_f.write(data)

        docker_id = host.connector_data["docker_container_id"]
        docker_command = "docker cp {0} {1}:{2}".format(
            temp_filename,
            docker_id,
            remote_filename,
        )

        status, _, stderr = run_local_shell_command(
            state,
            host,
            docker_command,
            print_output=print_output,
            print_input=print_input,
        )
    finally:
        os.close(fd)
        os.remove(temp_filename)

    if not status:
        raise IOError("\n".join(stderr))

    if print_output:
        click.echo(
            "{0}file uploaded to container: {1}".format(
                host.print_prefix,
                remote_filename,
            ),
            err=True,
        )

    return status


def get_file(
    state: "State",
    host: "Host",
    remote_filename,
    filename_or_io,
    remote_temp_filename=None,  # ignored
    print_output=False,
    print_input=False,
    **kwargs,  # ignored (sudo/etc)
):
    """
    Download a file from the target Docker container by copying it to a temporary
    location and then reading that into our final file/IO object.
    """

    fd, temp_filename = mkstemp()

    try:
        docker_id = host.connector_data["docker_container_id"]
        docker_command = "docker cp {0}:{1} {2}".format(
            docker_id,
            remote_filename,
            temp_filename,
        )

        status, _, stderr = run_local_shell_command(
            state,
            host,
            docker_command,
            print_output=print_output,
            print_input=print_input,
        )

        # Load the temporary file and write it to our file or IO object
        with open(temp_filename) as temp_f:
            with get_file_io(filename_or_io, "wb") as file_io:
                data = temp_f.read()
                data_bytes: bytes

                if isinstance(data, str):
                    data_bytes = data.encode()
                else:
                    data_bytes = data

                file_io.write(data_bytes)
    finally:
        os.close(fd)
        os.remove(temp_filename)

    if not status:
        raise IOError("\n".join(stderr))

    if print_output:
        click.echo(
            "{0}file downloaded from container: {1}".format(
                host.print_prefix,
                remote_filename,
            ),
            err=True,
        )

    return status



"""
Manage git repositories and configuration.
"""

import re

from pyinfra import host
from pyinfra.api import OperationError, operation
from pyinfra.facts.files import Directory, File
from pyinfra.facts.git import GitBranch, GitConfig, GitTrackingBranch

from . import files, ssh
from .util.files import chown, unix_path_join


@operation(
    pipeline_facts={"git_config": "repo"},
)
def config(key, value, repo=None):
    """
    Manage git config for a repository or globally.

    + key: the key of the config to ensure
    + value: the value this key should have
    + repo: specify the git repo path to edit local config (defaults to global)

    **Example:**

    .. code:: python

        git.config(
            name="Ensure user name is set for a repo",
            key="user.name",
            value="Anon E. Mouse",
            repo="/usr/local/src/pyinfra",
        )

    """

    existing_config = {}

    if not repo:
        existing_config = host.get_fact(GitConfig)

    # Only get the config if the repo exists at this stage
    elif host.get_fact(Directory, path=unix_path_join(repo, ".git")):
        existing_config = host.get_fact(GitConfig, repo=repo)

    if existing_config.get(key) != value:
        if repo is None:
            yield 'git config --global {0} "{1}"'.format(key, value)
        else:
            yield 'cd {0} && git config --local {1} "{2}"'.format(repo, key, value)

        existing_config[key] = value

    else:
        host.noop("git config {0} is set to {1}".format(key, value))


@operation(
    pipeline_facts={"git_branch": "target"},
)
def repo(
    src,
    dest,
    branch=None,
    pull=True,
    rebase=False,
    user=None,
    group=None,
    ssh_keyscan=False,
    update_submodules=False,
    recursive_submodules=False,
):
    """
    Clone/pull git repositories.

    + src: the git source URL
    + dest: directory to clone to
    + branch: branch to pull/checkout
    + pull: pull any changes for the branch
    + rebase: when pulling, use ``--rebase``
    + user: chown files to this user after
    + group: chown files to this group after
    + ssh_keyscan: keyscan the remote host if not in known_hosts before clone/pull
    + update_submodules: update any git submodules
    + recursive_submodules: update git submodules recursively

    **Example:**

    .. code:: python

        git.repo(
            name="Clone repo",
            src="https://github.com/Fizzadar/pyinfra.git",
            dest="/usr/local/src/pyinfra",
        )
    """

    # Ensure our target directory exists
    yield from files.directory(dest)

    # Do we need to scan for the remote host key?
    if ssh_keyscan:
        # Attempt to parse the domain from the git repository
        domain = re.match(r"^[a-zA-Z0-9]+@([0-9a-zA-Z\.\-]+)", src)

        if domain:
            yield from ssh.keyscan(domain.group(1))
        else:
            raise OperationError(
                "Could not parse domain (to SSH keyscan) from: {0}".format(src),
            )

    # Store git commands for directory prefix
    git_commands = []
    git_dir = unix_path_join(dest, ".git")
    is_repo = host.get_fact(Directory, path=git_dir)

    # Cloning new repo?
    if not is_repo:
        if branch:
            git_commands.append("clone {0} --branch {1} .".format(src, branch))
        else:
            git_commands.append("clone {0} .".format(src))

        host.create_fact(GitBranch, kwargs={"repo": dest}, data=branch)
        host.create_fact(
            Directory,
            kwargs={"path": git_dir},
            data={"user": user, "group": group},
        )

    # Ensuring existing repo
    else:
        if branch and host.get_fact(GitBranch, repo=dest) != branch:
            git_commands.append("fetch")  # fetch to ensure we have the branch locally
            git_commands.append("checkout {0}".format(branch))
            host.create_fact(GitBranch, kwargs={"repo": dest}, data=branch)

        if pull:
            if rebase:
                git_commands.append("pull --rebase")
            else:
                git_commands.append("pull")

    if update_submodules:
        if recursive_submodules:
            git_commands.append("submodule update --init --recursive")
        else:
            git_commands.append("submodule update --init")

    # Attach prefixes for directory
    command_prefix = "cd {0} && git".format(dest)
    git_commands = ["{0} {1}".format(command_prefix, command) for command in git_commands]

    for cmd in git_commands:
        yield cmd

    # Apply any user or group if we did anything
    if git_commands and (user or group):
        yield chown(dest, user, group, recursive=True)


@operation()
def worktree(
    worktree,
    repo=None,
    detached=False,
    new_branch=None,
    commitish=None,
    pull=True,
    rebase=False,
    from_remote_branch=None,
    present=True,
    assume_repo_exists=False,
    force=False,
    user=None,
    group=None,
):
    """
    Manage git worktrees.

    + worktree: git working tree directory
    + repo: git main repository directory
    + detached: create a working tree with a detached HEAD
    + new_branch: local branch name created at the same time than the worktree
    + commitish: from which git commit, branch, ... the worktree is created
    + pull: pull any changes from a remote branch if set
    + rebase: when pulling, use ``--rebase``
    + from_remote_branch: a 2-tuple ``(remote, branch)`` that identifies a remote branch
    + present: whether the working tree should exist
    + assume_repo_exists: whether to assume the main repo exists
    + force: remove unclean working tree if should not exist
    + user: chown files to this user after
    + group: chown files to this group after

    **Examples:**

    .. code:: python

        git.worktree(
            name="Create a worktree from the current repo `HEAD`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix"
        )

        git.worktree(
            name="Create a worktree from the commit `4e091aa0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            commitish="4e091aa0"
        )

        git.worktree(
            name="Create a worktree with a new local branch `v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            new_branch="v1.0",
        )

        git.worktree(
            name="Create a worktree from the commit 4e091aa0 with the new local branch `v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            new_branch="v1.0",
            commitish="4e091aa0"
        )

        git.worktree(
            name="Create a worktree with a detached `HEAD`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            detached=True,
        )

        git.worktree(
            name="Create a worktree with a detached `HEAD` from commit `4e091aa0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            commitish="4e091aa0",
            detached=True,
        )

        git.worktree(
            name="Create a worktree from the existing local branch `v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            commitish="v1.0"
        )

        git.worktree(
            name="Create a worktree with a new branch `v1.0` that tracks `origin/v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            new_branch="v1.0",
            commitish="v1.0"
        )

        git.worktree(
            name="Pull an existing worktree already linked to a tracking branch",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix"
        )

        git.worktree(
            name="Pull an existing worktree from a specific remote branch",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            from_remote_branch=("origin", "master")
        )

        git.worktree(
            name="Remove a worktree",
            worktree="/usr/local/src/pyinfra/hotfix",
            present=False,
        )

        git.worktree(
            name="Remove an unclean worktree",
            worktree="/usr/local/src/pyinfra/hotfix",
            present=False,
            force=True,
        )
    """

    # Doesn't exist & we want it
    if not host.get_fact(Directory, path=worktree) and present:

        # be sure that `repo` is a GIT repository
        if not assume_repo_exists and not host.get_fact(
            Directory,
            path=unix_path_join(repo, ".git"),
        ):
            raise OperationError(
                "The following folder is not a valid GIT repository : {0}".format(repo),
            )

        command_parts = ["cd {0} && git worktree add".format(repo)]

        if new_branch:
            command_parts.append("-b {0}".format(new_branch))
        elif detached:
            command_parts.append("--detach")

        command_parts.append(worktree)

        if commitish:
            command_parts.append(commitish)

        yield " ".join(command_parts)

        # Apply any user or group
        if user or group:
            yield chown(worktree, user, group, recursive=True)

        host.create_fact(Directory, kwargs={"path": worktree}, data={"user": user, "group": group})
        host.create_fact(GitTrackingBranch, kwargs={"repo": worktree}, data=new_branch)

    # It exists and we don't want it
    elif host.get_fact(Directory, path=worktree) and not present:

        command = "cd {0} && git worktree remove .".format(worktree)

        if force:
            command += " --force"

        yield command

        host.delete_fact(Directory, kwargs={"path": worktree})
        host.create_fact(GitTrackingBranch, kwargs={"repo": worktree})

    # It exists and we still want it => pull/rebase it
    elif host.get_fact(Directory, path=worktree) and present:

        # pull the worktree only if it's already linked to a tracking branch or
        # if a remote branch is set
        if host.get_fact(GitTrackingBranch, repo=worktree) or from_remote_branch:
            command = "cd {0} && git pull".format(worktree)

            if rebase:
                command += " --rebase"

            if from_remote_branch:
                if len(from_remote_branch) != 2 or type(from_remote_branch) not in (tuple, list):
                    raise OperationError(
                        "The remote branch must be a 2-tuple (remote, branch) such as "
                        '("origin", "master")',
                    )
                command += " {0} {1}".format(*from_remote_branch)

            yield command


@operation
def bare_repo(
    path,
    user=None,
    group=None,
    present=True,
):
    """
    Create bare git repositories.

    + path: path to the folder
    + present: whether the bare repository should exist
    + user: chown files to this user after
    + group: chown files to this group after

    **Example:**

    .. code:: python

        git.bare_repo(
            name="Create bare repo",
            path="/home/git/test.git",
        )
    """

    yield from files.directory(path, present=present)

    if present:
        head_filename = unix_path_join(path, "HEAD")
        head_file = host.get_fact(File, path=head_filename)

        if not head_file:
            yield "git init --bare {0}".format(path)
            if user or group:
                yield chown(path, user, group, recursive=True)
        else:
            if (user and head_file["user"] != user) or (group and head_file["group"] != group):
                yield chown(path, user, group, recursive=True)

        host.create_fact(
            File,
            kwargs={"path": head_filename},
            data={"user": user, "group": group, "mode": None},
        )



import re

from pyinfra.api import FactBase

# Valid unit names consist of a "name prefix" and a dot and a suffix specifying the unit type.
# The "unit prefix" must consist of one or more valid characters
# (ASCII letters, digits, ":", "-", "_", ".", and "\").
# The total length of the unit name including the suffix must not exceed 256 characters.
# The type suffix must be one of
# ".service", ".socket", ".device", ".mount", ".automount",
# ".swap", ".target", ".path", ".timer", ".slice", or ".scope".
# Units names can be parameterized by a single argument called the "instance name".
# A template unit must have a single "@" at the end of the name (right before the type suffix).
# The name of the full unit is formed by inserting the instance name
# between "@" and the unit type suffix.
SYSTEMD_UNIT_NAME_REGEX = (
    r"[a-zA-Z0-9\:\-\_\.\\\@]+\."
    r"(?:service|socket|device|mount|automount|swap|target|path|timer|slice|scope)"
)


def _make_systemctl_cmd(user_mode=False, machine=None, user_name=None):
    # base command for normal and user mode
    systemctl_cmd = "systemctl --user" if user_mode else "systemctl"

    # add user and machine flag if given in args
    if machine is not None:
        if user_name is not None:
            machine_opt = "--machine={1}@{0}".format(machine, user_name)
        else:
            machine_opt = "--machine={0}".format(machine)

        systemctl_cmd = "{0} {1}".format(systemctl_cmd, machine_opt)

    return systemctl_cmd


class SystemdStatus(FactBase):
    """
    Returns a dictionary map of systemd units to booleans indicating whether they are active.

    + user_mode: whether to use user mode
    + machine: machine name

    .. code:: python

        {
            "ssh.service": True,
            "containerd.service": True,
            "apt-daily.timer": False,
        }
    """

    requires_command = "systemctl"

    default = dict

    state_key = "SubState"
    state_values = ["running", "waiting", "exited"]

    def command(self, user_mode=False, machine=None, user_name=None):
        fact_cmd = _make_systemctl_cmd(
            user_mode=user_mode,
            machine=machine,
            user_name=user_name,
        )

        return f"{fact_cmd} show --all --property Id --property {self.state_key} '*'"

    def process(self, output):
        services = {}

        current_unit = None
        for line in output:
            line = line.strip()

            try:
                key, value = line.split("=", 1)
            except ValueError:
                current_unit = None  # reset current_unit just in case
                continue

            if key == "Id" and re.match(SYSTEMD_UNIT_NAME_REGEX, value):
                current_unit = value
                continue

            if key == self.state_key and current_unit:
                services[current_unit] = value in self.state_values

        return services


class SystemdEnabled(SystemdStatus):
    """
    Returns a dictionary map of systemd units to booleans indicating whether they are enabled.

    .. code:: python

        {
            "ssh.service": True,
            "containerd.service": True,
            "apt-daily.timer": False,
        }
    """

    state_key = "UnitFileState"
    state_values = ["enabled", "static"]



"""
The iptables modules handles iptables rules
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.api.exceptions import OperationError
from pyinfra.facts.iptables import Ip6tablesChains, Ip6tablesRules, IptablesChains, IptablesRules


@operation
def chain(
    chain,
    present=True,
    table="filter",
    policy=None,
    version=4,
):
    """
    Add/remove/update iptables chains.

    + chain: the name of the chain
    + present: whether the chain should exist
    + table: the iptables table this chain should belong to
    + policy: the policy this table should have
    + version: whether to target iptables or ip6tables

    Policy:
        These can only be applied to system chains (FORWARD, INPUT, OUTPUT, etc).
    """

    chains = (
        host.get_fact(IptablesChains, table=table)
        if version == 4
        else host.get_fact(Ip6tablesChains, table=table)
    )

    command = "iptables" if version == 4 else "ip6tables"
    command = "{0} -t {1}".format(command, table)

    if not present:
        if chain in chains:
            yield "{0} -X {1}".format(command, chain)
            chains.pop(chain)
        else:
            host.noop("iptables chain {0} does not exist".format(chain))
        return

    if present:
        if chain not in chains:
            yield "{0} -N {1}".format(command, chain)
            chains[chain] = None  # policy will be set below
        else:
            host.noop("iptables chain {0} exists".format(chain))

        if policy:
            if chain not in chains or chains[chain] != policy:
                yield "{0} -P {1} {2}".format(command, chain, policy)
                chains[chain] = policy


@operation
def rule(
    chain,
    jump,
    present=True,
    table="filter",
    append=True,
    version=4,
    # Core iptables filter arguments
    protocol=None,
    not_protocol=None,
    source=None,
    not_source=None,
    destination=None,
    not_destination=None,
    in_interface=None,
    not_in_interface=None,
    out_interface=None,
    not_out_interface=None,
    # After-rule arguments
    to_destination=None,
    to_source=None,
    to_ports=None,
    log_prefix=None,
    # Extras and extra shortcuts
    destination_port=None,
    source_port=None,
    extras="",
):
    """
    Add/remove iptables rules.

    + chain: the chain this rule should live in
    + jump: the target of the rule
    + table: the iptables table this rule should belong to
    + append: whether to append or insert the rule (if not present)
    + version: whether to target iptables or ip6tables

    Iptables args:

    + protocol/not_protocol: filter by protocol (tcp or udp)
    + source/not_source: filter by source IPs
    + destination/not_destination: filter by destination IPs
    + in_interface/not_in_interface: filter by incoming interface
    + out_interface/not_out_interface: filter by outgoing interface
    + to_destination: where to route to when jump=DNAT
    + to_source: where to route to when jump=SNAT
    + to_ports: where to route to when jump=REDIRECT
    + log_prefix: prefix for the log of this rule when jump=LOG

    Extras:

    + extras: a place to define iptables extension arguments (eg --limit, --physdev)
    + destination_port: destination port (requires protocol)
    + source_port: source port (requires protocol)

    **Examples:**

    .. code:: python

        iptables.rule(
            name="Block SSH traffic",
            chain="INPUT",
            jump="DROP",
            destination_port=22,
        )

        iptables.rule(
            name="NAT traffic on from 8.8.8.8:53 to 8.8.4.4:8080",
            chain="PREROUTING",
            jump="DNAT",
            table="nat",
            source="8.8.8.8",
            destination_port=53,
            to_destination="8.8.4.4:8080",
        )
    """

    if isinstance(to_ports, int):
        to_ports = "{0}".format(to_ports)

    # These are only shortcuts for extras
    if destination_port:
        extras = "{0} --dport {1}".format(extras, destination_port)

    if source_port:
        extras = "{0} --sport {1}".format(extras, source_port)

    # Convert the extras string into a set to enable comparison with the fact
    extras_set = set(extras.split())

    # When protocol is set, the extension is automagically added by iptables (which shows
    # in iptables-save): http://ipset.netfilter.org/iptables-extensions.man.html
    if protocol:
        extras_set.add("-m")
        extras_set.add(protocol)

    # --dport and --sport do not work without a protocol (because they need -m [tcp|udp]
    elif destination_port or source_port:
        raise OperationError(
            "iptables cannot filter by destination_port/source_port without a protocol",
        )

    # Verify NAT arguments, --to-destination only w/table=nat, jump=DNAT
    if to_destination and (table != "nat" or jump != "DNAT"):
        raise OperationError(
            "iptables only supports to_destination on the nat table and the DNAT jump "
            "(table={0}, jump={1})".format(table, jump),
        )

    # As above, --to-source only w/table=nat, jump=SNAT
    if to_source and (table != "nat" or jump != "SNAT"):
        raise OperationError(
            "iptables only supports to_source on the nat table and the SNAT jump "
            "(table={0}, jump={1})".format(table, jump),
        )

    # As above, --to-ports only w/table=nat, jump=REDIRECT
    if to_ports and (table != "nat" or jump != "REDIRECT"):
        raise OperationError(
            "iptables only supports to_ports on the nat table and the REDIRECT jump "
            "(table={0}, jump={1})".format(table, jump),
        )

    # --log-prefix is only supported with jump=LOG
    if log_prefix and jump != "LOG":
        raise OperationError(
            "iptables only supports log_prefix with the LOG jump " "(jump={0})".format(jump),
        )

    definition = {
        "chain": chain,
        "jump": jump,
        "protocol": protocol,
        "source": source,
        "destination": destination,
        "in_interface": in_interface,
        "out_interface": out_interface,
        "not_protocol": not_protocol,
        "not_source": not_source,
        "not_destination": not_destination,
        "not_in_interface": not_in_interface,
        "not_out_interface": not_out_interface,
        # These go *after* the jump argument
        "log_prefix": log_prefix,
        "to_destination": to_destination,
        "to_source": to_source,
        "to_ports": to_ports,
        "extras": extras_set,
    }

    definition = {
        key: (
            "{0}/32".format(value)
            if (
                key in ("source", "not_source", "destination", "not_destination")
                and "/" not in value
            )
            else value
        )
        for key, value in definition.items()
        if value
    }

    rules = (
        host.get_fact(IptablesRules, table=table)
        if version == 4
        else host.get_fact(Ip6tablesRules, table=table)
    )

    action = None

    # Definition doesn't exist and we want it
    if present:
        if definition not in rules:
            action = "-A" if append else "-I"
        else:
            host.noop("iptables {0} rule exists".format(chain))
            return

    # Definition exists and we don't want it
    if not present:
        if definition in rules:
            action = "-D"
        else:
            host.noop("iptables {0} rule does not exists".format(chain))
            return

    # Are we adding/removing a rule? Lets build it
    if action:
        args = [
            "iptables" if version == 4 else "ip6tables",
            # Add the table
            "-t",
            table,
            # Add the action and target chain
            action,
            chain,
        ]

        if protocol:
            args.extend(("-p", protocol))

        if source:
            args.extend(("-s", source))

        if in_interface:
            args.extend(("-i", in_interface))

        if out_interface:
            args.extend(("-o", out_interface))

        if not_protocol:
            args.extend(("!", "-p", not_protocol))

        if not_source:
            args.extend(("!", "-s", not_source))

        if not_in_interface:
            args.extend(("!", "-i", not_in_interface))

        if not_out_interface:
            args.extend(("!", "-o", not_out_interface))

        if extras:
            args.append(extras.strip())

        # Add the jump
        args.extend(("-j", jump))

        if log_prefix:
            args.extend(("--log-prefix", log_prefix))

        if to_destination:
            args.extend(("--to-destination", to_destination))

        if to_source:
            args.extend(("--to-source", to_source))

        if to_ports:
            args.extend(("--to-ports", to_ports))

        # Build the final iptables command
        yield " ".join(args)

        if action == "-D":
            rules.remove(definition)
        else:
            rules.append(definition)



import re
from datetime import datetime

WIN_LS_REGEX = re.compile(
    (
        # filetype and mode
        r"^([darhsl\-]{6})\s+"
        # Windows date
        r"([0-9]{1,2}\/[0-9]{1,2}\/[0-9]{4})\s+([0-9]{1,2}:[0-9]{1,2}\s[AP][M])\s+"
        # Size (Note: no size on directories)
        r"([0-9]+)?\s+"
        # Size and Filename
        r"([\w\/\.@-]+)"
    ),
)

WIN_FLAG_TO_TYPE = {
    "d": "directory",
    "-": "file",
}

WIN_FLAG_TO_ATTR = {
    "a": "archive",
    "r": "readonly",
    "h": "hidden",
    "s": "system",
    "l": "link",
    "-": "none",
}


def _parse_time(time):
    # Try matching windows format
    try:
        tmp = datetime.strptime(time, "%m/%d/%Y %H:%M %p")
        return tmp
    except ValueError:
        pass


def parse_win_ls_output(output, wanted_type):
    if output:
        matches = re.match(WIN_LS_REGEX, output)
        if matches:
            if output[5] == "l":
                type = "link"
            else:
                type = WIN_FLAG_TO_TYPE[output[0]]

            if type != wanted_type:
                return False

            hidden = False
            system = False
            readonly = False
            archive = False
            link = False

            mode_without_first = matches.group(1)[1:]

            for c in mode_without_first:
                tmp = WIN_FLAG_TO_ATTR[c]
                if tmp != "none":
                    if tmp == "hidden":
                        hidden = True
                    if tmp == "readonly":
                        readonly = True
                    if tmp == "archive":
                        archive = True
                    if tmp == "system":
                        system = True
                    if tmp == "link":
                        link = True

            mode = {
                "archive": archive,
                "hidden": hidden,
                "readonly": readonly,
                "system": system,
                "link": link,
            }

            date_and_time = "{} {}".format(matches.group(2), matches.group(3))

            size = "0"
            if type == "file":
                size = matches.group(4)

            out = {
                "type": type,
                "mode": mode,
                "mtime": _parse_time(date_and_time),
                "size": size,
                "name": matches.group(5),
                # TODO: You will need to run another powershell command to
                #       get the link target, so bailing on that for now.
            }

            return out



import re
from collections import defaultdict

from pyinfra.api import FactBase, MaskString, QuoteString, StringCommand
from pyinfra.api.util import try_int

from .util.databases import parse_columns_and_rows


def make_mysql_command(
    database=None,
    user=None,
    password=None,
    host=None,
    port=None,
    executable="mysql",
):
    target_bits = [executable]

    if database:
        target_bits.append(database)

    if user:
        # Quote the username as in may contain special characters
        target_bits.append('-u"{0}"'.format(user))

    if password:
        # Quote the password as it may contain special characters
        target_bits.append(MaskString('-p"{0}"'.format(password)))

    if host:
        target_bits.append("-h{0}".format(host))

    if port:
        target_bits.append("-P{0}".format(port))

    return StringCommand(*target_bits)


def make_execute_mysql_command(command, ignore_errors=False, **mysql_kwargs):
    commands_bits = [
        make_mysql_command(**mysql_kwargs),
        "-Be",
        QuoteString(command),  # quote this whole item as a single shell argument
    ]

    if ignore_errors:
        commands_bits.extend(["||", "true"])

    return StringCommand(*commands_bits)


class MysqlFactBase(FactBase):
    abstract = True

    requires_command = "mysql"
    ignore_errors = False

    def command(
        self,
        # Details for speaking to MySQL via `mysql` CLI via `mysql` CLI
        mysql_user=None,
        mysql_password=None,
        mysql_host=None,
        mysql_port=None,
    ):
        return make_execute_mysql_command(
            self.mysql_command,
            ignore_errors=self.ignore_errors,
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )


class MysqlDatabases(MysqlFactBase):
    """
    Returns a dict of existing MySQL databases and associated data:

    .. code:: python

        {
            "mysql": {
                "character_set": "latin1",
                "collation_name": "latin1_swedish_ci"
            },
        }
    """

    default = dict
    mysql_command = "SELECT * FROM information_schema.SCHEMATA"

    def process(self, output):
        rows = parse_columns_and_rows(
            output,
            "\t",
            title_parser=lambda title: title.lower(),
        )

        databases = {}

        for details in rows:
            databases[details.pop("schema_name")] = {
                "character_set": details["default_character_set_name"],
                "collation_name": details["default_collation_name"],
            }

        return databases


class MysqlUsers(MysqlFactBase):
    """
    Returns a dict of MySQL ``user@host``'s and their associated data:

    .. code:: python

        {
            "user@host": {
                "privileges": ["Alter", "Grant"],
                'max_connections': 5,
                ...
            },
        }
    """

    default = dict
    mysql_command = "SELECT * FROM mysql.user"

    @staticmethod
    def process(output):
        rows = parse_columns_and_rows(output, "\t")

        users = {}

        for details in rows:
            if details.get("Host") is None or details.get("User") is None:
                continue  # pragma: no cover

            privileges = []

            for key, value in list(details.items()):
                if key.endswith("_priv") and details.pop(key) == "Y":
                    privileges.append(key.replace("_priv", ""))

                if key.startswith("max_"):
                    details[key] = try_int(value)

                if key in ("password_expired", "is_role"):
                    details[key] = value == "Y"

            details["privileges"] = sorted(privileges)

            # Attach the user in the format user@host
            users[
                "{0}@{1}".format(
                    details.pop("User"),
                    details.pop("Host"),
                )
            ] = details

        return users


MYSQL_GRANT_REGEX = (
    r"^GRANT ([A-Z,\s]+) ON ((?:\*|`[a-z_\\]+`)(?:\.\*|'[a-z_]+')) "
    r"TO `[A-Z0-9a-z_\-]+`@`[A-Z0-9a-z_\.\-]+`(.*)"
)


class MysqlUserGrants(MysqlFactBase):
    """
    Returns a dict of ``<database>`.<table>`` with a set of granted privileges for each:

    .. code:: python

        {
            "`pyinfra_stuff`.*": {
                "SELECT",
                "INSERT",
                "GRANT OPTION",
            },
        }
    """

    default = dict
    # Ignore errors as SHOW GRANTS will error if the user does not exist
    ignore_errors = True

    def command(
        self,
        user,
        hostname="localhost",
        # Details for speaking to MySQL via `mysql` CLI via `mysql` CLI
        mysql_user=None,
        mysql_password=None,
        mysql_host=None,
        mysql_port=None,
    ):
        self.mysql_command = 'SHOW GRANTS FOR "{0}"@"{1}"'.format(user, hostname)

        return super(MysqlUserGrants, self).command(
            mysql_user,
            mysql_password,
            mysql_host,
            mysql_port,
        )

    @staticmethod
    def process(output):
        database_table_privileges = defaultdict(set)

        for line in output:
            matches = re.match(MYSQL_GRANT_REGEX, line)
            if not matches:
                continue

            privileges, database_table, extras = matches.groups()

            # MySQL outputs this pre-escaped
            database_table = database_table.replace("\\\\", "\\")

            for privilege in privileges.split(","):
                privilege = privilege.strip()
                database_table_privileges[database_table].add(privilege)

            if "WITH GRANT OPTION" in extras:
                database_table_privileges[database_table].add("GRANT OPTION")

        return database_table_privileges



import re
from collections import defaultdict

from pyinfra.api import FactBase

FIELDS = ["user", "role", "type", "level"]  # order is significant, do not change


class SEBoolean(FactBase):
    """
    Returns the status of a SELinux Boolean as a string (``on`` or ``off``).
    If ``boolean`` does not exist, ``SEBoolean`` returns the empty string.
    """

    requires_command = "getsebool"
    default = str

    def command(self, boolean):
        return "getsebool {0}".format(boolean)

    def process(self, output):
        components = output[0].split(" --> ")
        return components[1]


class FileContext(FactBase):
    """
    Returns structured SELinux file context data for a specified file
    or ``None`` if the file does not exist.

    .. code:: python

        {
            "user": "system_u",
            "role": "object_r",
            "type": "default_t",
            "level": "s0",
        }
    """

    def command(self, path):
        return "stat -c %C {0} || exit 0".format(path)

    def process(self, output):
        context = {}
        components = output[0].split(":")
        context["user"] = components[0]
        context["role"] = components[1]
        context["type"] = components[2]
        context["level"] = components[3]
        return context


class FileContextMapping(FactBase):
    """
    Returns structured SELinux file context data for the specified target path prefix
    using the same format as :ref:`selinux.FileContext`.  If there is no mapping, it returns ``{}``
    Note: This fact requires root privileges.
    """

    requires_command = "semanage"
    default = dict

    def command(self, target):
        return "set -o pipefail && semanage fcontext -n -l | (grep '^{0}' || true)".format(target)

    def process(self, output):
        # example output: /etc       all files          system_u:object_r:etc_t:s0 # noqa: SC100
        # but lines at end that won't match: /etc/systemd/system = /usr/lib/systemd/system
        if len(output) != 1:
            return self.default()
        m = re.match(r"^.*\s+(\w+):(\w+):(\w+):(\w+)", output[0])
        return {k: m.group(i) for i, k in enumerate(FIELDS, 1)} if m is not None else self.default()


class SEPorts(FactBase):
    """
    Returns the SELinux 'type' definitions for ``(tcp|udp|dccp|sctp)`` ports.
    Note: This fact requires root privileges.

    .. code:: python
        {
            "tcp": { 22: "ssh_port_t", ...},
            "udp": { ...}
        }
    """

    requires_command = "semanage"
    default = dict
    # example output: amqp_port_t                    tcp      15672, 5671-5672  # noqa: SC100
    _regex = re.compile(r"^([\w_]+)\s+(\w+)\s+([\w\-,\s]+)$")

    def command(self):
        return "semanage port -ln"

    def process(self, output):
        labels = defaultdict(dict)
        for line in output:
            m = SEPorts._regex.match(line)
            if m is None:  # something went wrong
                continue
            if m.group(1) == "unreserved_port_t":  # these cover the entire space
                continue
            for item in m.group(3).split(","):
                item = item.strip()
                if "-" in item:
                    pieces = item.split("-")
                    start, stop = int(pieces[0]), int(pieces[1])
                else:
                    start = stop = int(item)
                labels[m.group(2)].update({port: m.group(1) for port in range(start, stop + 1)})

        return labels


class SEPort(FactBase):
    """
    Returns the SELinux 'type' for the specified protocol ``(tcp|udp|dccp|sctp)`` and port number.
    If no type has been set, ``SEPort`` returns the empty string.
    Note: ``policycoreutils-dev`` must be installed for this to work.
    """

    requires_command = "sepolicy"
    default = str

    def command(self, protocol, port):
        return "(sepolicy network -p {0} 2>/dev/null || true) | grep {1}".format(port, protocol)

    def process(self, output):
        # if type set, first line is specific and second is generic type for port range
        # each rows in the format "22: tcp ssh_port_t 22"

        return output[0].split(" ")[2] if len(output) > 1 else self.default()



from unittest import TestCase
from unittest.mock import MagicMock, patch

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all

from ..util import make_inventory


class TestWinrmConnector(TestCase):
    def setUp(self):
        self.fake_connect_patch = patch("pyinfra.connectors.winrm.connect")
        self.fake_connect_mock = self.fake_connect_patch.start()

    def tearDown(self):
        self.fake_connect_patch.stop()

    def test_connect_host(self):
        inventory = make_inventory(hosts=[("@winrm/somehost", {})])
        state = State(inventory, Config())
        host = inventory.get_host("@winrm/somehost")
        host.connect(reason=True)
        assert len(state.active_hosts) == 0
        assert host.data.winrm_hostname == "somehost"

    def test_connect_all_password(self):
        inventory = make_inventory(
            hosts=(
                ("@winrm/somehost", {"winrm_username": "testuser", "winrm_password": "testpass"}),
                (
                    "@winrm/anotherhost",
                    {"winrm_username": "testuser2", "winrm_password": "testpass2"},
                ),
            ),
        )
        state = State(inventory, Config())
        connect_all(state)

        # Get a host
        somehost = inventory.get_host("@winrm/somehost")
        assert somehost.data.winrm_username == "testuser"
        assert somehost.data.winrm_password == "testpass"

        assert len(state.active_hosts) == 2

    def test_run_shell_command(self):
        fake_winrm_session = MagicMock()

        fake_winrm = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_winrm.run_cmd.return_value = fake_stdin, fake_stdout, MagicMock()

        fake_winrm_session.return_value = fake_winrm

        inventory = make_inventory(hosts=("@winrm/somehost",))
        State(inventory, Config())
        host = inventory.get_host("@winrm/somehost")
        host.connect()

        command = "echo hi"

        out = host.run_shell_command(command, stdin="hello", print_output=True)
        assert len(out) == 3

        status, stdout, stderr = out
        # TODO: assert status is True

        combined_out = host.run_shell_command(
            command,
            print_output=True,
            return_combined_output=True,
        )
        assert len(combined_out) == 2

        # TODO: fake_winrm.run_shell_command.assert_called_with("'echo hi'", get_pty=False)



import re

from pyinfra.api import FactBase

from .util.packaging import parse_packages

rpm_regex = r"^(\S+)\ (\S+)$"
rpm_query_format = "%{NAME} %{VERSION}-%{RELEASE}\\n"


class RpmPackages(FactBase):
    """
    Returns a dict of installed rpm packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = 'rpm --queryformat "{0}" -qa'.format(rpm_query_format)
    requires_command = "rpm"

    default = dict

    def process(self, output):
        return parse_packages(rpm_regex, output)


class RpmPackage(FactBase):
    """
    Returns information on a .rpm file:

    .. code:: python

        {
            "name": "my_package",
            "version": "1.0.0",
        }
    """

    requires_command = "rpm"

    def command(self, name):
        return (
            'rpm --queryformat "{0}" -q {1} || '
            "! test -e {1} || "
            'rpm --queryformat "{0}" -qp {1} 2> /dev/null'
        ).format(rpm_query_format, name)

    def process(self, output):
        for line in output:
            matches = re.match(rpm_regex, line)
            if matches:
                return {
                    "name": matches.group(1),
                    "version": matches.group(2),
                }


class RpmPackageProvides(FactBase):
    """
    Returns a list of packages that provide the specified capability (command, file, etc).
    """

    default = list

    requires_command = "repoquery"

    @staticmethod
    def command(name):
        # Accept failure here (|| true) for invalid/unknown packages
        return 'repoquery --queryformat "{0}" --whatprovides {1} || true'.format(
            rpm_query_format,
            name,
        )

    @staticmethod
    def process(output):
        packages = []

        for line in output:
            matches = re.match(rpm_regex, line)
            if matches:
                packages.append(list(matches.groups()))

        return packages



from typing import TYPE_CHECKING

import gevent

from pyinfra.progress import progress_spinner

if TYPE_CHECKING:
    from pyinfra.api.state import State


def connect_all(state: "State"):
    """
    Connect to all the configured servers in parallel. Reads/writes state.inventory.

    Args:
        state (``pyinfra.api.State`` obj): the state containing an inventory to connect to
    """

    hosts = [
        host
        for host in state.inventory
        if state.is_host_in_limit(host)  # these are the hosts to activate ("initially connect to")
    ]

    greenlet_to_host = {state.pool.spawn(host.connect): host for host in hosts}

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            progress(host)

    # Get/set the results
    failed_hosts = set()

    for greenlet, host in greenlet_to_host.items():
        # Raise any unexpected exception
        greenlet.get()

        if host.connection:
            state.activate_host(host)
        else:
            failed_hosts.add(host)

    # Remove those that failed, triggering FAIL_PERCENT check
    state.fail_hosts(failed_hosts, activated_count=len(hosts))


def disconnect_all(state: "State"):
    for host in state.activated_hosts:  # only hosts we connected to please!
        host.disconnect()  # normally a noop



# encoding: utf-8

from socket import error as socket_error, gaierror
from unittest import TestCase
from unittest.mock import MagicMock, call, mock_open, patch

from paramiko import AuthenticationException, PasswordRequiredException, SSHException

import pyinfra
from pyinfra.api import Config, MaskString, State, StringCommand
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import ConnectError, PyinfraError
from pyinfra.connectors.ssh import _get_sftp_connection

from ..util import make_inventory


def make_raise_exception_function(cls, *args, **kwargs):
    def handler(*a, **kw):
        raise cls(*args, **kwargs)

    return handler


@patch("pyinfra.connectors.ssh.SSHClient.get_transport", MagicMock())
@patch("pyinfra.connectors.ssh.open", mock_open(read_data="test!"), create=True)
class TestSSHConnector(TestCase):
    def setUp(self):
        self.fake_connect_patch = patch("pyinfra.connectors.ssh.SSHClient.connect")
        self.fake_connect_mock = self.fake_connect_patch.start()
        _get_sftp_connection.cache = {}

    def tearDown(self):
        self.fake_connect_patch.stop()

    def test_connect_all(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)
        assert len(state.active_hosts) == 2

    def test_connect_host(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect(reason=True)
        assert len(state.active_hosts) == 0

    def test_connect_all_password(self):
        inventory = make_inventory(override_data={"ssh_password": "test"})

        # Get a host
        somehost = inventory.get_host("somehost")
        assert somehost.data.ssh_password == "test"

        state = State(inventory, Config())
        connect_all(state)

        assert len(state.active_hosts) == 2

    @patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True)
    @patch("pyinfra.connectors.ssh.RSAKey.from_private_key_file")
    def test_connect_exceptions(self, fake_key_open):
        for exception_class in (
            AuthenticationException,
            SSHException,
            gaierror,
            socket_error,
            EOFError,
        ):
            state = State(make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)), Config())

            self.fake_connect_mock.side_effect = make_raise_exception_function(exception_class)

            with self.assertRaises(PyinfraError):
                connect_all(state)

            assert len(state.active_hosts) == 0

    # SSH key tests
    #

    def test_connect_with_rsa_ssh_key(self):
        state = State(make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)), Config())

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_key_open:

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            connect_all(state)

            # Check the key was created properly
            fake_key_open.assert_called_with(filename="testkey")
            # Check the certificate file was then loaded
            fake_key.load_certificate.assert_called_with("testkey.pub")

            # And check the Paramiko SSH call was correct
            self.fake_connect_mock.assert_called_with(
                "somehost",
                allow_agent=False,
                look_for_keys=False,
                pkey=fake_key,
                timeout=10,
                username="vagrant",
                _pyinfra_ssh_forward_agent=None,
                _pyinfra_ssh_config_file=None,
                _pyinfra_ssh_known_hosts_file=None,
                _pyinfra_ssh_strict_host_key_checking=None,
                _pyinfra_ssh_paramiko_connect_kwargs=None,
            )

        # Check that loading the same key again is cached in the state
        second_state = State(
            make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)),
            Config(),
        )
        second_state.private_keys = state.private_keys

        connect_all(second_state)

    def test_connect_with_rsa_ssh_key_password(self):
        state = State(
            make_inventory(
                hosts=(("somehost", {"ssh_key": "testkey", "ssh_key_password": "testpass"}),),
            ),
            Config(),
        )

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_key_open:
            fake_key = MagicMock()

            def fake_key_open_fail(*args, **kwargs):
                if "password" not in kwargs:
                    raise PasswordRequiredException()
                return fake_key

            fake_key_open.side_effect = fake_key_open_fail

            connect_all(state)

            # Check the key was created properly
            fake_key_open.assert_called_with(filename="testkey", password="testpass")
            # Check the certificate file was then loaded
            fake_key.load_certificate.assert_called_with("testkey.pub")

    def test_connect_with_rsa_ssh_key_password_from_prompt(self):
        state = State(make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)), Config())

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.getpass",
            lambda *args, **kwargs: "testpass",
        ), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_key_open:
            fake_key = MagicMock()

            def fake_key_open_fail(*args, **kwargs):
                if "password" not in kwargs:
                    raise PasswordRequiredException()
                return fake_key

            fake_key_open.side_effect = fake_key_open_fail

            pyinfra.is_cli = True
            connect_all(state)
            pyinfra.is_cli = False

            # Check the key was created properly
            fake_key_open.assert_called_with(filename="testkey", password="testpass")
            # Check the certificate file was then loaded
            fake_key.load_certificate.assert_called_with("testkey.pub")

    def test_connect_with_rsa_ssh_key_missing_password(self):
        state = State(make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)), Config())

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_key_open:

            fake_key_open.side_effect = make_raise_exception_function(PasswordRequiredException)

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            with self.assertRaises(PyinfraError) as e:
                connect_all(state)

            assert e.exception.args[0] == (
                "Private key file (testkey) is encrypted, set ssh_key_password " "to use this key"
            )

    def test_connect_with_rsa_ssh_key_wrong_password(self):
        state = State(
            make_inventory(
                hosts=(("somehost", {"ssh_key": "testkey", "ssh_key_password": "testpass"}),),
            ),
            Config(),
        )

        fake_fail_from_private_key_file = MagicMock()
        fake_fail_from_private_key_file.side_effect = make_raise_exception_function(SSHException)

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.DSSKey.from_private_key_file",
            fake_fail_from_private_key_file,
        ), patch(
            "pyinfra.connectors.ssh.ECDSAKey.from_private_key_file",
            fake_fail_from_private_key_file,
        ), patch(
            "pyinfra.connectors.ssh.Ed25519Key.from_private_key_file",
            fake_fail_from_private_key_file,
        ), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_key_open:

            def fake_key_open_fail(*args, **kwargs):
                if "password" not in kwargs:
                    raise PasswordRequiredException
                raise SSHException

            fake_key_open.side_effect = fake_key_open_fail

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            with self.assertRaises(PyinfraError) as e:
                connect_all(state)

            assert e.exception.args[0] == "Invalid private key file: testkey"

        assert fake_fail_from_private_key_file.call_count == 3

    def test_connect_with_dss_ssh_key(self):
        state = State(make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)), Config())

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_rsa_key_open, patch(
            "pyinfra.connectors.ssh.DSSKey.from_private_key_file",
        ) as fake_key_open:  # noqa

            fake_rsa_key_open.side_effect = make_raise_exception_function(SSHException)

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            connect_all(state)

            # Check the key was created properly
            fake_key_open.assert_called_with(filename="testkey")

            # And check the Paramiko SSH call was correct
            self.fake_connect_mock.assert_called_with(
                "somehost",
                allow_agent=False,
                look_for_keys=False,
                pkey=fake_key,
                timeout=10,
                username="vagrant",
                _pyinfra_ssh_forward_agent=None,
                _pyinfra_ssh_config_file=None,
                _pyinfra_ssh_known_hosts_file=None,
                _pyinfra_ssh_strict_host_key_checking=None,
                _pyinfra_ssh_paramiko_connect_kwargs=None,
            )

        # Check that loading the same key again is cached in the state
        second_state = State(
            make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)),
            Config(),
        )
        second_state.private_keys = state.private_keys

        connect_all(second_state)

    def test_connect_with_dss_ssh_key_password(self):
        state = State(
            make_inventory(
                hosts=(("somehost", {"ssh_key": "testkey", "ssh_key_password": "testpass"}),),
            ),
            Config(),
        )

        with patch("pyinfra.connectors.ssh.path.isfile", lambda *args, **kwargs: True), patch(
            "pyinfra.connectors.ssh.RSAKey.from_private_key_file",
        ) as fake_rsa_key_open, patch(
            "pyinfra.connectors.ssh.DSSKey.from_private_key_file",
        ) as fake_dss_key_open:  # noqa

            def fake_rsa_key_open_fail(*args, **kwargs):
                if "password" not in kwargs:
                    raise PasswordRequiredException
                raise SSHException

            fake_rsa_key_open.side_effect = fake_rsa_key_open_fail

            fake_dss_key = MagicMock()

            def fake_dss_key_func(*args, **kwargs):
                if "password" not in kwargs:
                    raise PasswordRequiredException
                return fake_dss_key

            fake_dss_key_open.side_effect = fake_dss_key_func

            connect_all(state)

            # Check the key was created properly
            fake_dss_key_open.assert_called_with(filename="testkey", password="testpass")

            # And check the Paramiko SSH call was correct
            self.fake_connect_mock.assert_called_with(
                "somehost",
                allow_agent=False,
                look_for_keys=False,
                pkey=fake_dss_key,
                timeout=10,
                username="vagrant",
                _pyinfra_ssh_forward_agent=None,
                _pyinfra_ssh_config_file=None,
                _pyinfra_ssh_known_hosts_file=None,
                _pyinfra_ssh_strict_host_key_checking=None,
                _pyinfra_ssh_paramiko_connect_kwargs=None,
            )

        # Check that loading the same key again is cached in the state
        second_state = State(
            make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)),
            Config(),
        )
        second_state.private_keys = state.private_keys

        connect_all(second_state)

    def test_connect_with_missing_ssh_key(self):
        state = State(make_inventory(hosts=(("somehost", {"ssh_key": "testkey"}),)), Config())

        with self.assertRaises(PyinfraError) as e:
            connect_all(state)

        self.assertTrue(e.exception.args[0].startswith("No such private key file:"))

    # SSH command tests
    #

    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_run_shell_command(self, fake_ssh_client):
        fake_ssh = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = fake_stdin, fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        command = "echo Šablony"
        fake_stdout.channel.recv_exit_status.return_value = 0

        out = host.run_shell_command(command, stdin="hello", print_output=True)
        assert len(out) == 3

        status, stdout, stderr = out
        assert status is True
        fake_stdin.write.assert_called_with(b"hello\n")

        combined_out = host.run_shell_command(
            command,
            stdin="hello",
            print_output=True,
            return_combined_output=True,
        )
        assert len(combined_out) == 2

        fake_ssh.exec_command.assert_called_with("sh -c 'echo Šablony'", get_pty=False)

    @patch("pyinfra.connectors.ssh.click")
    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_run_shell_command_masked(self, fake_ssh_client, fake_click):
        fake_ssh = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = MagicMock(), fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        command = StringCommand("echo", MaskString("top-secret-stuff"))
        fake_stdout.channel.recv_exit_status.return_value = 0

        out = host.run_shell_command(command, print_output=True, print_input=True)
        assert len(out) == 3

        status, stdout, stderr = out
        assert status is True

        fake_ssh.exec_command.assert_called_with(
            "sh -c 'echo top-secret-stuff'",
            get_pty=False,
        )

        fake_click.echo.assert_called_with(
            "{0}>>> sh -c 'echo ***'".format(host.print_prefix),
            err=True,
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_run_shell_command_success_exit_code(self, fake_ssh_client):
        fake_ssh = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = MagicMock(), fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        command = "echo hi"
        fake_stdout.channel.recv_exit_status.return_value = 1

        out = host.run_shell_command(command, success_exit_codes=[1])
        assert len(out) == 3
        assert out[0] is True

    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_run_shell_command_error(self, fake_ssh_client):
        fake_ssh = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = MagicMock(), fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect(state)

        command = "echo hi"
        fake_stdout.channel.recv_exit_status.return_value = 1

        out = host.run_shell_command(command)
        assert len(out) == 3
        assert out[0] is False

    @patch("pyinfra.connectors.util.getpass")
    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_run_shell_command_sudo_password_prompt(
        self,
        fake_ssh_client,
        fake_getpass,
    ):
        fake_ssh = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = MagicMock(), fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh
        fake_getpass.return_value = "password"

        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        command = "echo Šablony"
        fake_stdout.__iter__.return_value = ["/tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX"]
        fake_stdout.channel.recv_exit_status.return_value = 0

        out = host.run_shell_command(command, sudo=True, use_sudo_password=True, print_output=True)
        assert len(out) == 3

        status, stdout, stderr = out
        assert status is True

        fake_ssh.exec_command.assert_called_with(
            (
                "env SUDO_ASKPASS=/tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX "
                "PYINFRA_SUDO_PASSWORD=password "
                "sudo -H -A -k sh -c 'echo Šablony'"
            ),
            get_pty=False,
        )

    @patch("pyinfra.connectors.util.getpass")
    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_run_shell_command_sudo_password_automatic_prompt(
        self,
        fake_ssh_client,
        fake_getpass,
    ):
        fake_ssh = MagicMock()
        first_fake_stdout = MagicMock()
        second_fake_stdout = MagicMock()
        third_fake_stdout = MagicMock()

        first_fake_stdout.__iter__.return_value = ["sudo: a password is required\r"]
        second_fake_stdout.__iter__.return_value = ["/tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX"]

        fake_ssh.exec_command.side_effect = [
            (MagicMock(), first_fake_stdout, MagicMock()),  # command w/o sudo password
            (MagicMock(), second_fake_stdout, MagicMock()),  # SUDO_ASKPASS_COMMAND
            (MagicMock(), third_fake_stdout, MagicMock()),  # command with sudo pw
        ]

        fake_ssh_client.return_value = fake_ssh
        fake_getpass.return_value = "password"

        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        command = "echo Šablony"
        first_fake_stdout.channel.recv_exit_status.return_value = 1
        second_fake_stdout.channel.recv_exit_status.return_value = 0
        third_fake_stdout.channel.recv_exit_status.return_value = 0

        out = host.run_shell_command(command, sudo=True, print_output=True)
        assert len(out) == 3

        status, stdout, stderr = out
        assert status is True

        fake_ssh.exec_command.assert_any_call(("sudo -H -n sh -c 'echo Šablony'"), get_pty=False)

        fake_ssh.exec_command.assert_called_with(
            (
                "env SUDO_ASKPASS=/tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX "
                "PYINFRA_SUDO_PASSWORD=password "
                "sudo -H -A -k sh -c 'echo Šablony'"
            ),
            get_pty=False,
        )

    # SSH file put/get tests
    #

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.util._get_sudo_password")
    def test_run_shell_command_retry_for_sudo_password(
        self,
        fake_get_sudo_password,
        fake_ssh_client,
    ):
        fake_get_sudo_password.return_value = "PASSWORD"

        fake_ssh = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_stderr = ["sudo: a password is required"]
        fake_ssh.exec_command.return_value = fake_stdin, fake_stdout, fake_stderr

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect(state)
        host.connector_data["sudo_askpass_path"] = "/tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX"

        command = "echo hi"
        return_values = [1, 0]  # return 0 on the second call
        fake_stdout.channel.recv_exit_status.side_effect = lambda: return_values.pop(0)

        out = host.run_shell_command(command)
        assert len(out) == 3
        assert out[0] is True
        assert fake_get_sudo_password.called
        fake_ssh.exec_command.assert_called_with(
            "env SUDO_ASKPASS=/tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX "
            "PYINFRA_SUDO_PASSWORD=PASSWORD sh -c 'echo hi'",
            get_pty=False,
        )

    # SSH file put/get tests
    #

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_put_file(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("anotherhost",))
        State(inventory, Config())
        host = inventory.get_host("anotherhost")
        host.connect()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.put_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
            )

        assert status is True
        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(),
            "not-another-file",
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_put_file_sudo(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("anotherhost",))
        State(inventory, Config())
        host = inventory.get_host("anotherhost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.put_file(
                "not-a-file",
                "not another file",
                print_output=True,
                sudo=True,
                sudo_user="ubuntu",
            )

        assert status is True

        fake_ssh_client().exec_command.assert_called_with(
            ("sh -c 'rm -f " "/tmp/pyinfra-de01e82cb691e8a31369da3c7c8f17341c44ac24'"),
            get_pty=False,
        )

        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(),
            "/tmp/pyinfra-de01e82cb691e8a31369da3c7c8f17341c44ac24",
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_put_file_su_user_fail_acl(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("anotherhost",))
        State(inventory, Config())
        host = inventory.get_host("anotherhost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 1
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.put_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
                su_user="centos",
            )

        assert status is False

        fake_ssh_client().exec_command.assert_called_with(
            (
                "sh -c 'setfacl -m u:centos:r "
                "/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487'"
            ),
            get_pty=False,
        )

        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(),
            "/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487",
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_put_file_su_user_fail_copy(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("anotherhost",))
        State(inventory, Config())
        host = inventory.get_host("anotherhost")
        host.connect()

        stdout_mock = MagicMock()
        exit_codes = [0, 1]
        stdout_mock.channel.recv_exit_status.side_effect = lambda: exit_codes.pop(0)
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.put_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
                su_user="centos",
            )

        assert status is False

        fake_ssh_client().exec_command.assert_any_call(
            (
                "sh -c 'setfacl -m u:centos:r "
                "/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487'"
            ),
            get_pty=False,
        )

        fake_ssh_client().exec_command.assert_called_with(
            (
                "su centos -c 'sh -c '\"'\"'cp "
                "/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487 "
                "not-another-file'\"'\"''"
            ),
            get_pty=False,
        )

        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(),
            "/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487",
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_put_file_sudo_custom_temp_file(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("anotherhost",))
        State(inventory, Config())
        host = inventory.get_host("anotherhost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.put_file(
                "not-a-file",
                "not another file",
                print_output=True,
                sudo=True,
                sudo_user="ubuntu",
                remote_temp_filename="/a-different-tempfile",
            )

        assert status is True

        fake_ssh_client().exec_command.assert_called_with(
            ("sh -c 'rm -f " "/a-different-tempfile'"),
            get_pty=False,
        )

        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(),
            "/a-different-tempfile",
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_get_file(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.get_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
            )

        assert status is True
        fake_sftp_client.from_transport().getfo.assert_called_with(
            "not-a-file",
            fake_open(),
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_get_file_sudo(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.get_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
                sudo=True,
                sudo_user="ubuntu",
            )

        assert status is True

        fake_ssh_client().exec_command.assert_has_calls(
            [
                call(
                    (
                        "sudo -H -n -u ubuntu sh -c 'cp not-a-file "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r not-a-file'"  # noqa
                    ),
                    get_pty=False,
                ),
                call(
                    (
                        "sudo -H -n -u ubuntu sh -c 'rm -f "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508'"
                    ),
                    get_pty=False,
                ),
            ],
        )

        fake_sftp_client.from_transport().getfo.assert_called_with(
            "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508",
            fake_open(),
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    def test_get_file_sudo_copy_fail(self, fake_ssh_client):
        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 1
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        status = host.get_file(
            "not-a-file",
            "not-another-file",
            print_output=True,
            sudo=True,
            sudo_user="ubuntu",
        )

        assert status is False

        fake_ssh_client().exec_command.assert_has_calls(
            [
                call(
                    (
                        "sudo -H -n -u ubuntu sh -c 'cp not-a-file "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r not-a-file'"  # noqa
                    ),
                    get_pty=False,
                ),
            ],
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_get_file_sudo_remove_fail(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.side_effect = [0, 1]
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.get_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
                sudo=True,
                sudo_user="ubuntu",
            )

        assert status is False

        fake_ssh_client().exec_command.assert_has_calls(
            [
                call(
                    (
                        "sudo -H -n -u ubuntu sh -c 'cp not-a-file "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r not-a-file'"  # noqa
                    ),
                    get_pty=False,
                ),
                call(
                    (
                        "sudo -H -n -u ubuntu sh -c 'rm -f "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508'"
                    ),
                    get_pty=False,
                ),
            ],
        )

        fake_sftp_client.from_transport().getfo.assert_called_with(
            "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508",
            fake_open(),
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_get_file_su_user(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("somehost",))
        State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            status = host.get_file(
                "not-a-file",
                "not-another-file",
                print_output=True,
                su_user="centos",
            )

        assert status is True

        fake_ssh_client().exec_command.assert_has_calls(
            [
                call(
                    (
                        "su centos -c 'sh -c '\"'\"'cp not-a-file "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r "
                        "not-a-file'\"'\"''"
                    ),
                    get_pty=False,
                ),
                call(
                    (
                        "su centos -c 'sh -c '\"'\"'rm -f "
                        "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508'\"'\"''"
                    ),
                    get_pty=False,
                ),
            ],
        )

        fake_sftp_client.from_transport().getfo.assert_called_with(
            "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508",
            fake_open(),
        )

    @patch("pyinfra.connectors.ssh.SSHClient")
    @patch("pyinfra.connectors.ssh.SFTPClient")
    def test_get_sftp_fail(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=("anotherhost",))
        State(inventory, Config())
        host = inventory.get_host("anotherhost")
        host.connect()

        fake_sftp_client.from_transport.side_effect = make_raise_exception_function(SSHException)

        fake_open = mock_open(read_data="test!")
        with patch("pyinfra.api.util.open", fake_open, create=True):
            with self.assertRaises(ConnectError):
                host.put_file(
                    "not-a-file",
                    "not-another-file",
                    print_output=True,
                )



"""
Manage BSD init services (``/etc/rc.d``, ``/usr/local/etc/rc.d``).
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.bsdinit import RcdStatus
from pyinfra.facts.server import Os

from . import files
from .util.service import handle_service_control


@operation
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
):
    """
    Manage the state of BSD init services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<service> <command>``
    + enabled: whether this service should be enabled/disabled on boot
    """

    status_argument = "status"
    if host.get_fact(Os) == "OpenBSD":
        status_argument = "check"

    yield from handle_service_control(
        host,
        service,
        host.get_fact(RcdStatus),
        "test -e /etc/rc.d/{0} && /etc/rc.d/{0} {1} || /usr/local/etc/rc.d/{0} {1}",
        running,
        restarted,
        reloaded,
        command,
        status_argument=status_argument,
    )

    # BSD init is simple, just add/remove <service>_enabled="YES"
    if isinstance(enabled, bool):
        yield from files.line(
            "/etc/rc.conf.local",
            "^{0}_enable=".format(service),
            replace='{0}_enable="YES"'.format(service),
            present=enabled,
        )



# encoding: utf-8

import shlex
from unittest import TestCase
from unittest.mock import MagicMock, mock_open, patch

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import InventoryError, PyinfraError
from pyinfra.connectors.util import make_unix_command

from ..util import make_inventory


def fake_ssh_docker_shell(
    state,
    host,
    command,
    get_pty=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    use_sudo_password=False,
    **command_kwargs,
):
    if host.data.ssh_hostname not in ("somehost", "anotherhost"):
        raise PyinfraError("Invalid host", host.data.ssh_hostname)

    if command == "docker run -d not-an-image tail -f /dev/null":
        return (True, ["containerid"], [])

    elif command == "docker commit containerid":
        return (True, ["sha256:blahsomerandomstringdata"], [])

    elif command == "docker rm -f containerid":
        return (True, [], [])

    elif str(command).startswith("rm -f"):
        return (True, [], [])

    # This is a bit messy. But it's easier than trying to swap out a mock
    # when it needs to be used...
    elif fake_ssh_docker_shell.custom_command:
        custom_command, status, stdout, stderr = fake_ssh_docker_shell.custom_command
        if str(command) == custom_command:
            fake_ssh_docker_shell.ran_custom_command = True
            return (status, stdout, stderr)

    raise PyinfraError("Invalid Command: {0}".format(command))


def get_docker_command(command):
    shell_command = make_unix_command(command).get_raw_value()
    shell_command = shlex.quote(shell_command)
    docker_command = "docker exec -it containerid sh -c {0}".format(shell_command)
    return docker_command


@patch("pyinfra.connectors.ssh.connect", MagicMock())
@patch("pyinfra.connectors.ssh.run_shell_command", fake_ssh_docker_shell)
@patch("pyinfra.api.util.open", mock_open(read_data="test!"), create=True)
class TestDockerSSHConnector(TestCase):
    def setUp(self):
        # reset custom command for shell
        fake_ssh_docker_shell.custom_command = None
        fake_ssh_docker_shell.ran_custom_command = False

    def test_connect_host(self):
        inventory = make_inventory(("somehost", "anotherhost"))
        state = State(inventory, Config())
        host = inventory.get_host("somehost")
        host.connect(reason=True)
        assert len(state.active_hosts) == 0

    def test_missing_image_and_host(self):
        with self.assertRaises(InventoryError):
            make_inventory(hosts=("@dockerssh",))

    def test_missing_image(self):
        with self.assertRaises(InventoryError):
            make_inventory(hosts=("@dockerssh/host",))

        with self.assertRaises(InventoryError):
            make_inventory(hosts=("@dockerssh/host:",))

    def test_connect_all(self):
        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())
        connect_all(state)
        assert len(state.active_hosts) == 1

    def test_user_provided_container_id(self):
        inventory = make_inventory(
            hosts=(("@dockerssh/somehost:not-an-image", {"docker_container_id": "abc"}),),
        )
        State(inventory, Config())
        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect()
        assert host.data.docker_container_id == "abc"

    def test_connect_all_error(self):
        inventory = make_inventory(hosts=("@dockerssh/somehost:a-broken-image",))
        state = State(inventory, Config())

        with self.assertRaises(PyinfraError):
            connect_all(state)

    def test_connect_disconnect_host(self):
        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())
        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect(reason=True)
        assert len(state.active_hosts) == 0
        host.disconnect()

    def test_run_shell_command(self):
        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        State(inventory, Config())

        command = "echo hi"

        fake_ssh_docker_shell.custom_command = [get_docker_command(command), True, [], []]

        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect()
        out = host.run_shell_command(
            command,
            stdin="hello",
            get_pty=True,
            print_output=True,
        )
        assert len(out) == 3
        assert out[0] is True
        assert fake_ssh_docker_shell.ran_custom_command

    def test_run_shell_command_error(self):
        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())

        command = "echo hi"
        fake_ssh_docker_shell.custom_command = [get_docker_command(command), False, [], []]

        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect(state)
        out = host.run_shell_command(command, get_pty=True)
        assert out[0] is False
        assert fake_ssh_docker_shell.ran_custom_command

    @patch("pyinfra.connectors.dockerssh.mkstemp", lambda: (None, "local_tempfile"))
    @patch("pyinfra.connectors.docker.os.close", lambda f: None)
    @patch("pyinfra.connectors.dockerssh.ssh.put_file")
    def test_put_file(self, fake_put_file):
        fake_ssh_docker_shell.custom_command = [
            "docker cp remote_tempfile containerid:not-another-file",
            True,
            [],
            [],
        ]

        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: "remote_tempfile"

        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect()

        host.put_file("not-a-file", "not-another-file", print_output=True)

        # ensure copy from local to remote host
        fake_put_file.assert_called_with(state, host, "local_tempfile", "remote_tempfile")

        # ensure copy from remote host to remote docker container
        assert fake_ssh_docker_shell.ran_custom_command

    @patch("pyinfra.connectors.dockerssh.ssh.put_file")
    def test_put_file_error(self, fake_put_file):
        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: "remote_tempfile"

        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect()

        # SSH error
        fake_put_file.return_value = False
        with self.assertRaises(IOError):
            host.put_file("not-a-file", "not-another-file", print_output=True)

        # docker copy error
        fake_ssh_docker_shell.custom_command = [
            "docker cp remote_tempfile containerid:not-another-file",
            False,
            [],
            ["docker error"],
        ]
        fake_put_file.return_value = True

        with self.assertRaises(IOError) as e:
            host.put_file("not-a-file", "not-another-file", print_output=True)
        assert str(e.exception) == "docker error"

    @patch("pyinfra.connectors.dockerssh.ssh.get_file")
    def test_get_file(self, fake_get_file):
        fake_ssh_docker_shell.custom_command = [
            "docker cp containerid:not-a-file remote_tempfile",
            True,
            [],
            [],
        ]

        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: "remote_tempfile"

        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect()

        host.get_file("not-a-file", "not-another-file", print_output=True)

        # ensure copy from local to remote host
        fake_get_file.assert_called_with(state, host, "remote_tempfile", "not-another-file")

        # ensure copy from remote host to remote docker container
        assert fake_ssh_docker_shell.ran_custom_command

    @patch("pyinfra.connectors.dockerssh.ssh.get_file")
    def test_get_file_error(self, fake_get_file):
        fake_ssh_docker_shell.custom_command = [
            "docker cp containerid:not-a-file remote_tempfile",
            False,
            [],
            ["docker error"],
        ]

        inventory = make_inventory(hosts=("@dockerssh/somehost:not-an-image",))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: "remote_tempfile"

        host = inventory.get_host("@dockerssh/somehost:not-an-image")
        host.connect()

        fake_get_file.return_value = True
        with self.assertRaises(IOError) as ex:
            host.get_file("not-a-file", "not-another-file", print_output=True)

        assert str(ex.exception) == "docker error"

        # SSH error
        fake_ssh_docker_shell.custom_command = [
            "docker cp containerid:not-a-file remote_tempfile",
            True,
            [],
            [],
        ]
        fake_get_file.return_value = False
        with self.assertRaises(IOError) as ex:
            host.get_file("not-a-file", "not-another-file", print_output=True)

        assert str(ex.exception) == "failed to copy file over ssh"



from functools import wraps
from hashlib import sha1
from inspect import getframeinfo, getfullargspec, stack
from os import getcwd, path, stat
from socket import error as socket_error, timeout as timeout_error
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

import click
from jinja2 import Environment, FileSystemLoader, StrictUndefined
from paramiko import SSHException

import pyinfra
from pyinfra import logger

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State

# 64kb chunks
BLOCKSIZE = 65536

# Caches
TEMPLATES: Dict[Any, Any] = {}
FILE_SHAS: Dict[Any, Any] = {}

PYINFRA_API_DIR = path.dirname(__file__)


def get_file_path(state: "State", filename: str):
    if path.isabs(filename):
        return filename

    assert state.cwd is not None, "Cannot use `get_file_path` with no `state.cwd` set"
    relative_to = state.cwd

    if state.current_exec_filename and (filename.startswith("./") or filename.startswith(".\\")):
        relative_to = path.dirname(state.current_exec_filename)

    return path.join(relative_to, filename)


def get_args_kwargs_spec(func: Callable[..., Any]) -> Tuple[List, Dict]:
    args: List[Any] = []
    kwargs: Dict[Any, Any] = {}

    argspec = getfullargspec(func)
    if not argspec.args:
        return args, kwargs

    if argspec.defaults:
        kwargs = dict(
            zip(
                argspec.args[-len(argspec.defaults) :],
                argspec.defaults,
            ),
        )
        args = argspec.args[: -len(argspec.defaults)]
    else:
        args = argspec.args

    return args, kwargs


def get_kwargs_str(kwargs: Dict[Any, Any]):
    if not kwargs:
        return ""

    items = [
        "{0}={1}".format(key, value)
        for key, value in sorted(kwargs.items())
        if key not in ("self", "state", "host")
    ]
    return ", ".join(items)


def try_int(value):
    try:
        return int(value)
    except (TypeError, ValueError):
        return value


def memoize(func: Callable[..., Any]):
    @wraps(func)
    def wrapper(*args, **kwargs):
        key = "{0}{1}".format(args, kwargs)
        if key in wrapper.cache:
            return wrapper.cache[key]

        value = func(*args, **kwargs)
        wrapper.cache[key] = value
        return value

    wrapper.cache = {}  # type: ignore
    return wrapper


def get_call_location(frame_offset: int = 1):
    frame = get_caller_frameinfo(frame_offset=frame_offset)  # escape *this* function
    relpath = frame.filename

    try:
        # On Windows if pyinfra is on a different drive to the filename here, this will
        # error as there's no way to do relative paths between drives.
        relpath = path.relpath(frame.filename)
    except ValueError:
        pass

    return "line {0} in {1}".format(frame.lineno, relpath)


def get_caller_frameinfo(frame_offset: int = 0):
    # Default frames to look at is 2; one for this function call itself
    # in util.py and one for the caller of this function within pyinfra
    # giving the external call frame (ie end user deploy code).
    frame_shift = 2 + frame_offset

    stack_items = stack()

    frame = stack_items[frame_shift][0]
    info = getframeinfo(frame)

    del stack_items

    return info


def get_operation_order_from_stack(state: "State"):
    stack_items = list(reversed(stack()))

    # Find the *first* occurrence of our deploy file in the reversed stack
    if state.current_deploy_filename:
        for i, stack_item in enumerate(stack_items):
            frame = getframeinfo(stack_item[0])
            if frame.filename == state.current_deploy_filename:
                break
    else:
        i = 0

    # Now generate a list of line numbers *following that file*
    line_numbers = []

    if pyinfra.is_cli:
        line_numbers.append(state.current_op_file_number)

    for stack_item in stack_items[i:]:
        frame = getframeinfo(stack_item[0])

        if frame.filename.startswith(PYINFRA_API_DIR):
            continue

        line_numbers.append(frame.lineno)

    del stack_items

    return line_numbers


def get_template(filename_or_io: str):
    """
    Gets a jinja2 ``Template`` object for the input filename or string, with caching
    based on the filename of the template, or the SHA1 of the input string.
    """

    file_data = get_file_io(filename_or_io, mode="r")
    cache_key = file_data.cache_key

    if cache_key and cache_key in TEMPLATES:
        return TEMPLATES[cache_key]

    with file_data as file_io:
        template_string = file_io.read()

    template = Environment(
        undefined=StrictUndefined,
        keep_trailing_newline=True,
        loader=FileSystemLoader(getcwd()),
    ).from_string(template_string)

    if cache_key:
        TEMPLATES[cache_key] = template

    return template


def sha1_hash(string: str):
    """
    Return the SHA1 of the input string.
    """

    hasher = sha1()
    hasher.update(string.encode("utf-8"))
    return hasher.hexdigest()


def format_exception(e):
    return "{0}{1}".format(e.__class__.__name__, e.args)


def print_host_combined_output(host: "Host", combined_output_lines):
    for type_, line in combined_output_lines:
        if type_ == "stderr":
            logger.error(
                "{0}{1}".format(
                    host.print_prefix,
                    click.style(line, "red"),
                ),
            )
        else:
            logger.error(
                "{0}{1}".format(
                    host.print_prefix,
                    line,
                ),
            )


def log_operation_start(op_meta: Dict, op_types: Optional[List] = None, prefix: str = "--> "):
    op_types = op_types or []
    if op_meta["serial"]:
        op_types.append("serial")
    if op_meta["run_once"]:
        op_types.append("run once")

    args = ""
    if op_meta["args"]:
        args = "({0})".format(", ".join(str(arg) for arg in op_meta["args"]))

    logger.info(
        "{0} {1} {2}".format(
            click.style(
                "{0}Starting{1}operation:".format(
                    prefix,
                    " {0} ".format(", ".join(op_types)) if op_types else " ",
                ),
                "blue",
            ),
            click.style(", ".join(op_meta["names"]), bold=True),
            args,
        ),
    )


def log_error_or_warning(
    host: "Host", ignore_errors: bool, description: str = "", continue_on_error: bool = False
):
    log_func = logger.error
    log_color = "red"
    log_text = "Error: " if description else "Error"

    if ignore_errors:
        log_func = logger.warning
        log_color = "yellow"
        log_text = (
            "Error (ignored, execution continued)" if continue_on_error else "Error (ignored)"
        )
        if description:
            log_text = f"{log_text}: "

    log_func(
        "{0}{1}{2}".format(
            host.print_prefix,
            click.style(log_text, log_color),
            description,
        ),
    )


def log_host_command_error(host: "Host", e, timeout: int = 0):
    if isinstance(e, timeout_error):
        logger.error(
            "{0}{1}".format(
                host.print_prefix,
                click.style(
                    "Command timed out after {0}s".format(
                        timeout,
                    ),
                    "red",
                ),
            ),
        )

    elif isinstance(e, (socket_error, SSHException)):
        logger.error(
            "{0}{1}".format(
                host.print_prefix,
                click.style(
                    "Command socket/SSH error: {0}".format(format_exception(e)),
                    "red",
                ),
            ),
        )

    elif isinstance(e, IOError):
        logger.error(
            "{0}{1}".format(
                host.print_prefix,
                click.style(
                    "Command IO error: {0}".format(format_exception(e)),
                    "red",
                ),
            ),
        )

    # Still here? Re-raise!
    else:
        raise e


def make_hash(obj):
    """
    Make a hash from an arbitrary nested dictionary, list, tuple or set, used to generate
    ID's for operations based on their name & arguments.
    """

    if isinstance(obj, (set, tuple, list)):
        hash_string = "".join([make_hash(e) for e in obj])

    elif isinstance(obj, dict):
        hash_string = "".join("".join((key, make_hash(value))) for key, value in obj.items())

    else:
        hash_string = (
            # Capture integers first (as 1 == True)
            "{0}".format(obj)
            if isinstance(obj, int)
            # Constants - the values can change between hosts but we should still
            # group them under the same operation hash.
            else "_PYINFRA_CONSTANT"
            if obj in (True, False, None)
            # Plain strings
            else obj
            if isinstance(obj, str)
            # Objects with __name__s
            else obj.__name__
            if hasattr(obj, "__name__")
            # Objects with names
            else obj.name
            if hasattr(obj, "name")
            # Repr anything else
            else repr(obj)
        )

    return sha1_hash(hash_string)


class get_file_io(object):
    """
    Given either a filename or an existing IO object, this context processor
    will open and close filenames, and leave IO objects alone.
    """

    close = False

    def __init__(self, filename_or_io, mode="rb"):
        if not (
            # Check we can be read
            hasattr(filename_or_io, "read")
            # Or we're a filename
            or isinstance(filename_or_io, str)
        ):
            raise TypeError(
                "Invalid filename or IO object: {0}".format(
                    filename_or_io,
                ),
            )

        self.filename_or_io = filename_or_io
        self.mode = mode

    def __enter__(self):
        # If we have a read attribute, just use the object as-is
        if hasattr(self.filename_or_io, "read"):
            file_io = self.filename_or_io

        # Otherwise, assume a filename and open it up
        else:
            file_io = open(self.filename_or_io, self.mode)

            # Attach to self for closing on __exit__
            self.file_io = file_io
            self.close = True

        # Ensure we're at the start of the file
        file_io.seek(0)
        return file_io

    def __exit__(self, type, value, traceback):
        if self.close:
            self.file_io.close()

    @property
    def cache_key(self):
        # If we're a filename, cache against that - we don't cache in-memory
        # file objects.
        if isinstance(self.filename_or_io, str):
            return self.filename_or_io


def get_file_sha1(filename_or_io):
    """
    Calculates the SHA1 of a file or file object using a buffer to handle larger files.
    """

    file_data = get_file_io(filename_or_io)
    cache_key = file_data.cache_key

    if cache_key and cache_key in FILE_SHAS:
        return FILE_SHAS[cache_key]

    with file_data as file_io:
        hasher = sha1()
        buff = file_io.read(BLOCKSIZE)

        while len(buff) > 0:
            if isinstance(buff, str):
                buff = buff.encode("utf-8")

            hasher.update(buff)
            buff = file_io.read(BLOCKSIZE)

    digest = hasher.hexdigest()

    if cache_key:
        FILE_SHAS[cache_key] = digest

    return digest


def get_path_permissions_mode(pathname: str):
    """
    Get the permissions (bits) of a path as an integer.
    """

    mode_octal = oct(stat(pathname).st_mode)
    return mode_octal[-3:]



"""
Operations are the core of pyinfra. The ``@operation`` wrapper intercepts calls
to the function and instead diff against the remote server, outputting commands
to the deploy state. This is then run later by pyinfra's ``__main__`` or the
:doc:`./pyinfra.api.operations` module.
"""

from functools import wraps
from types import FunctionType
from typing import TYPE_CHECKING

import pyinfra
from pyinfra import context, logger
from pyinfra.context import ctx_host, ctx_state

from .arguments import get_execution_kwarg_keys, pop_global_arguments
from .command import StringCommand
from .exceptions import OperationValueError, PyinfraError
from .host import Host
from .operations import run_host_op
from .util import (
    get_args_kwargs_spec,
    get_call_location,
    get_operation_order_from_stack,
    log_operation_start,
    make_hash,
    memoize,
)

if TYPE_CHECKING:
    from pyinfra.api.state import State

op_meta_default = object()


class OperationMeta(object):
    combined_output_lines = None

    def __init__(self, hash=None, commands=None):
        # Wrap all the attributes
        commands = commands or []
        self.commands = commands
        self.hash = hash

        # Changed flag = did we do anything?
        self.changed = len(self.commands) > 0

    def __repr__(self):
        """
        Return Operation object as a string.
        """

        return (
            f"OperationMeta(commands={len(self.commands)}, "
            f"changed={self.changed}, hash={self.hash})"
        )

    def set_combined_output_lines(self, combined_output_lines):
        self.combined_output_lines = combined_output_lines

    def _get_lines(self, types=("stdout", "stderr")):
        if self.combined_output_lines is None:
            raise AttributeError("Output is not available until operations have been executed")

        return [line for type_, line in self.combined_output_lines if type_ in types]

    @property
    def stdout_lines(self):
        return self._get_lines(types=("stdout",))

    @property
    def stderr_lines(self):
        return self._get_lines(types=("stderr",))

    @property
    def stdout(self):
        return "\n".join(self.stdout_lines)

    @property
    def stderr(self):
        return "\n".join(self.stderr_lines)


def add_op(state: "State", op_func, *args, **kwargs):
    """
    Prepare & add an operation to ``pyinfra.state`` by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        to op_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    """

    allow_cli_mode = kwargs.pop("_allow_cli_mode", False)
    after_host_callback = kwargs.pop("_after_host_callback", lambda host: None)

    if pyinfra.is_cli and not allow_cli_mode:
        raise PyinfraError(
            ("`add_op` should not be called when pyinfra is executing in CLI mode! ({0})").format(
                get_call_location(),
            ),
        )

    hosts = kwargs.pop("host", state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    with ctx_state.use(state):
        results = {}
        for op_host in hosts:
            with ctx_host.use(op_host):
                results[op_host] = op_func(*args, **kwargs)
            after_host_callback(op_host)

    return results


@memoize
def show_state_host_arguments_warning(call_location):
    logger.warning(
        (
            "{0}:\n\tLegacy operation function detected! Operations should no longer define "
            "`state` and `host` arguments."
        ).format(call_location),
    )


def operation(
    func=None,
    pipeline_facts=None,
    is_idempotent: bool = True,
    idempotent_notice=None,
    frame_offset: int = 1,
):
    """
    Decorator that takes a simple module function and turn it into the internal
    operation representation that consists of a list of commands + options
    (sudo, (sudo|su)_user, env).
    """

    # If not decorating, return function with config attached
    if func is None:

        def decorator(f):
            f.pipeline_facts = pipeline_facts
            f.is_idempotent = is_idempotent
            f.idempotent_notice = idempotent_notice
            return operation(f, frame_offset=2)

        return decorator

    # Check whether an operation is "legacy" - ie contains state=None, host=None kwargs
    # TODO: remove this in v3
    is_legacy = False
    args, kwargs = get_args_kwargs_spec(func)
    if all(key in kwargs and kwargs[key] is None for key in ("state", "host")):
        show_state_host_arguments_warning(get_call_location(frame_offset=frame_offset))
        is_legacy = True
    func.is_legacy = is_legacy

    # Actually decorate!
    @wraps(func)
    def decorated_func(*args, **kwargs):
        state = context.state
        host = context.host

        # Configure operation
        #
        # Get the meta kwargs (globals that apply to all hosts)
        global_kwargs, global_kwarg_keys = pop_global_arguments(kwargs)

        # If this op is being called inside another, just return here
        # (any unwanted/op-related kwargs removed above).
        if host.in_op:
            if global_kwarg_keys:
                _error_msg = "Nested operation called with global arguments: {0} ({1})".format(
                    global_kwarg_keys,
                    get_call_location(),
                )
                raise PyinfraError(_error_msg)
            return func(*args, **kwargs) or []

        kwargs = _solve_legacy_operation_arguments(func, state, host, kwargs)
        names, add_args = _generate_operation_name(func, host, kwargs, global_kwargs)
        op_order, op_hash = _solve_operation_consistency(names, state, host)

        # Ensure shared (between servers) operation meta, mutates state
        op_meta = _ensure_shared_op_meta(state, op_hash, op_order, global_kwargs, names)

        # Attach normal args, if we're auto-naming this operation
        if add_args:
            op_meta = _attach_args(op_meta, args, kwargs)

        # Check if we're actually running the operation on this host
        # Run once and we've already added meta for this op? Stop here.
        if op_meta["run_once"]:
            has_run = False
            for ops in state.ops.values():
                if op_hash in ops:
                    has_run = True
                    break

            if has_run:
                return OperationMeta(op_hash)

        # "Run" operation
        #

        # Otherwise, flag as in-op and run it to get the commands
        host.in_op = True
        host.current_op_hash = op_hash
        host.current_op_global_kwargs = global_kwargs

        # Convert to list as the result may be a generator
        commands = func(*args, **kwargs)
        commands = [  # convert any strings -> StringCommand's
            StringCommand(command.strip()) if isinstance(command, str) else command
            for command in commands
        ]

        host.in_op = False
        host.current_op_hash = None
        host.current_op_global_kwargs = None

        # Add host-specific operation data to state, this mutates state
        operation_meta = _update_state_meta(state, host, commands, op_hash, op_meta, global_kwargs)

        # Return result meta for use in deploy scripts
        return operation_meta

    decorated_func._pyinfra_op = func  # type: ignore
    return decorated_func


def _solve_legacy_operation_arguments(op_func, state, host, kwargs):
    """
    Solve legacy operation arguments.
    """

    # If this is a legacy operation function (ie - state & host arg kwargs), ensure that state
    # and host are included as kwargs.

    # Legacy operation arguments
    if op_func.is_legacy:
        if "state" not in kwargs:
            kwargs["state"] = state
        if "host" not in kwargs:
            kwargs["host"] = host
    # If not legacy, pop off any state/host kwargs that may come from legacy @deploy functions
    else:
        kwargs.pop("state", None)
        kwargs.pop("host", None)

    return kwargs


def _generate_operation_name(func, host, kwargs, global_kwargs):
    # Generate an operation name if needed (Module/Operation format)
    name = global_kwargs.get("name")
    add_args = False
    if name:
        names = {name}
    else:
        add_args = True

        if func.__module__:
            module_bits = func.__module__.split(".")
            module_name = module_bits[-1]
            name = "{0}/{1}".format(module_name.title(), func.__name__.title())
        else:
            name = func.__name__

        names = {name}

    if host.current_deploy_name:
        names = {"{0} | {1}".format(host.current_deploy_name, name) for name in names}

    return names, add_args


def _solve_operation_consistency(names, state, host):
    # Operation order is used to tie-break available nodes in the operation DAG, in CLI mode
    # we use stack call order so this matches as defined by the user deploy code.
    if pyinfra.is_cli:
        op_order = get_operation_order_from_stack(state)
    # In API mode we just increase the order for each host
    else:
        op_order = [len(host.op_hash_order)]

    if host.loop_position:
        op_order.extend(host.loop_position)

    # Make a hash from the call stack lines
    op_hash = make_hash(op_order)

    # Avoid adding duplicates! This happens if an operation is called within
    # a loop - such that the filename/lineno/code _are_ the same, but the
    # arguments might be different. We just append an increasing number to
    # the op hash and also handle below with the op order.
    duplicate_op_count = 0
    while op_hash in host.op_hash_order:
        logger.debug("Duplicate hash ({0}) detected!".format(op_hash))
        op_hash = "{0}-{1}".format(op_hash, duplicate_op_count)
        duplicate_op_count += 1

    host.op_hash_order.append(op_hash)
    if duplicate_op_count:
        op_order.append(duplicate_op_count)

    op_order = tuple(op_order)
    logger.debug(f"Adding operation, {names}, opOrder={op_order}, opHash={op_hash}")
    return op_order, op_hash


# NOTE: this function mutates state.op_meta for this hash
def _ensure_shared_op_meta(state, op_hash, op_order, global_kwargs, names):
    op_meta = state.op_meta.setdefault(
        op_hash,
        {
            "names": set(),
            "args": [],
            "op_order": op_order,
        },
    )

    for key in get_execution_kwarg_keys():
        global_value = global_kwargs.pop(key)
        op_meta_value = op_meta.get(key, op_meta_default)

        if op_meta_value is not op_meta_default and global_value != op_meta_value:
            raise OperationValueError("Cannot have different values for `{0}`.".format(key))

        op_meta[key] = global_value

    # Add any new names to the set
    op_meta["names"].update(names)

    return op_meta


def _execute_immediately(state, host, op_data, op_meta, op_hash):
    logger.warning(
        f"Note: nested operations are currently in beta ({get_call_location()})\n"
        "    More information: "
        "https://docs.pyinfra.com/en/2.x/using-operations.html#nested-operations",
    )
    op_data["parent_op_hash"] = host.executing_op_hash
    log_operation_start(op_meta, op_types=["nested"], prefix="")
    run_host_op(state, host, op_hash)


def _attach_args(op_meta, args, kwargs):
    for arg in args:
        if isinstance(arg, FunctionType):
            arg = arg.__name__

        if arg not in op_meta["args"]:
            op_meta["args"].append(arg)

    # Attach keyword args
    for key, value in kwargs.items():
        if isinstance(value, FunctionType):
            value = value.__name__

        arg = "=".join((str(key), str(value)))
        if arg not in op_meta["args"]:
            op_meta["args"].append(arg)

    return op_meta


# NOTE: this function mutates state.meta for this host
def _update_state_meta(state, host, commands, op_hash, op_meta, global_kwargs):
    # We're doing some commands, meta/ops++
    state.meta[host]["ops"] += 1
    state.meta[host]["commands"] += len(commands)

    if commands:
        state.meta[host]["ops_change"] += 1
    else:
        state.meta[host]["ops_no_change"] += 1

    operation_meta = OperationMeta(op_hash, commands)

    # Add the server-relevant commands
    op_data = {
        "commands": commands,
        "global_kwargs": global_kwargs,
        "operation_meta": operation_meta,
    }
    state.set_op_data(host, op_hash, op_data)

    # If we're already in the execution phase, execute this operation immediately
    if state.is_executing:
        _execute_immediately(state, host, op_data, op_meta, op_hash)

    return operation_meta



from pyinfra.api import FactBase, MaskString, QuoteString, StringCommand
from pyinfra.api.util import try_int

from .util.databases import parse_columns_and_rows


def make_psql_command(
    database=None,
    user=None,
    password=None,
    host=None,
    port=None,
    executable="psql",
):
    target_bits = []

    if password:
        target_bits.append(MaskString('PGPASSWORD="{0}"'.format(password)))

    target_bits.append(executable)

    if database:
        target_bits.append("-d {0}".format(database))

    if user:
        target_bits.append("-U {0}".format(user))

    if host:
        target_bits.append("-h {0}".format(host))

    if port:
        target_bits.append("-p {0}".format(port))

    return StringCommand(*target_bits)


def make_execute_psql_command(command, **psql_kwargs):
    return StringCommand(
        make_psql_command(**psql_kwargs),
        "-Ac",
        QuoteString(command),  # quote this whole item as a single shell argument
    )


class PostgresqlFactBase(FactBase):
    abstract = True

    requires_command = "psql"

    def command(
        self,
        psql_user=None,
        psql_password=None,
        psql_host=None,
        psql_port=None,
    ):
        return make_execute_psql_command(
            self.psql_command,
            user=psql_user,
            password=psql_password,
            host=psql_host,
            port=psql_port,
        )


class PostgresqlRoles(PostgresqlFactBase):
    """
    Returns a dict of PostgreSQL roles and data:

    .. code:: python

        {
            "pyinfra": {
                "super": true,
                "createrole": false,
                "createdb": false,
                ...
            },
        }
    """

    default = dict
    psql_command = "SELECT * FROM pg_catalog.pg_roles"

    def process(self, output):
        # Remove the last line of the output (row count)
        output = output[:-1]
        rows = parse_columns_and_rows(
            output,
            "|",
            # Remove the "rol" prefix on column names
            remove_column_prefix="rol",
        )

        users = {}

        for details in rows:
            for key, value in list(details.items()):
                if key in ("oid", "connlimit"):
                    details[key] = try_int(value)

                if key in (
                    "super",
                    "inherit",
                    "createrole",
                    "createdb",
                    "canlogin",
                    "replication",
                    "bypassrls",
                ):
                    details[key] = value == "t"

            users[details.pop("name")] = details

        return users


class PostgresqlDatabases(PostgresqlFactBase):
    """
    Returns a dict of PostgreSQL databases and metadata:

    .. code:: python

        {
            "pyinfra_stuff": {
                "encoding": "UTF8",
                "collate": "en_US.UTF-8",
                "ctype": "en_US.UTF-8",
                ...
            },
        }
    """

    default = dict
    psql_command = "SELECT pg_catalog.pg_encoding_to_char(encoding), * FROM pg_catalog.pg_database"

    def process(self, output):
        # Remove the last line of the output (row count)
        output = output[:-1]
        rows = parse_columns_and_rows(
            output,
            "|",
            # Remove the "dat" prefix on column names
            remove_column_prefix="dat",
        )

        databases = {}

        for details in rows:
            details["encoding"] = details.pop("pg_encoding_to_char")

            for key, value in list(details.items()):
                if key.endswith("id") or key in (
                    "dba",
                    "tablespace",
                    "connlimit",
                ):
                    details[key] = try_int(value)

                if key in ("istemplate", "allowconn"):
                    details[key] = value == "t"

            databases[details.pop("name")] = details

        return databases



import shlex
from inspect import getfullargspec
from string import Formatter
from typing import TYPE_CHECKING

import gevent

from pyinfra.context import ctx_config, ctx_host

from .arguments import get_executor_kwarg_keys

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


def make_formatted_string_command(string: str, *args, **kwargs):
    """
    Helper function that takes a shell command or script as a string, splits it
    using ``shlex.split`` and then formats each bit, returning a ``StringCommand``
    instance with each bit.

    Useful to enable string formatted commands/scripts, for example:

    .. code:: python

        curl_command = make_formatted_string_command(
            'curl -sSLf {0} -o {1}',
            QuoteString(src),
            QuoteString(dest),
        )
    """

    formatter = Formatter()
    string_bits = []

    for bit in shlex.split(string):
        for item in formatter.parse(bit):
            if item[0]:
                string_bits.append(item[0])
            if item[1]:
                value, _ = formatter.get_field(item[1], args, kwargs)
                string_bits.append(value)

    return StringCommand(*string_bits)


class MaskString(str):
    pass


class QuoteString(object):
    def __init__(self, obj):
        self.object = obj

    def __repr__(self):
        return "QuoteString({0})".format(self.object)


class PyinfraCommand(object):
    def __init__(self, *args, **kwargs):
        self.executor_kwargs = {
            key: kwargs[key] for key in get_executor_kwarg_keys() if key in kwargs
        }

    def __eq__(self, other):
        if isinstance(other, self.__class__) and repr(self) == repr(other):
            return True
        return False

    def execute(self, state: "State", host: "Host", executor_kwargs):
        raise NotImplementedError


class StringCommand(PyinfraCommand):
    def __init__(self, *bits, **kwargs):
        super(StringCommand, self).__init__(**kwargs)
        self.bits = bits
        self.separator = kwargs.pop("_separator", " ")

    def __str__(self):
        return self.get_masked_value()

    def __repr__(self):
        return "StringCommand({0})".format(self.get_masked_value())

    def _get_all_bits(self, bit_accessor):
        all_bits = []

        for bit in self.bits:
            quote = False
            if isinstance(bit, QuoteString):
                quote = True
                bit = bit.object

            if isinstance(bit, StringCommand):
                bit = bit_accessor(bit)

            if not isinstance(bit, str):
                bit = "{0}".format(bit)

            if quote:
                bit = shlex.quote(bit)

            all_bits.append(bit)

        return all_bits

    def get_raw_value(self):
        return self.separator.join(
            self._get_all_bits(
                lambda bit: bit.get_raw_value(),
            ),
        )

    def get_masked_value(self):
        return self.separator.join(
            [
                "***" if isinstance(bit, MaskString) else bit
                for bit in self._get_all_bits(lambda bit: bit.get_masked_value())
            ],
        )

    def execute(self, state: "State", host: "Host", executor_kwargs):
        executor_kwargs.update(self.executor_kwargs)

        return host.run_shell_command(
            self,
            print_output=state.print_output,
            print_input=state.print_input,
            return_combined_output=True,
            **executor_kwargs,
        )


class FileUploadCommand(PyinfraCommand):
    def __init__(self, src: str, dest: str, remote_temp_filename=None, **kwargs):
        super(FileUploadCommand, self).__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.remote_temp_filename = remote_temp_filename

    def __repr__(self):
        return "FileUploadCommand({0}, {1})".format(self.src, self.dest)

    def execute(self, state: "State", host: "Host", executor_kwargs):
        executor_kwargs.update(self.executor_kwargs)

        return host.put_file(
            self.src,
            self.dest,
            remote_temp_filename=self.remote_temp_filename,
            print_output=state.print_output,
            print_input=state.print_input,
            **executor_kwargs,
        )


class FileDownloadCommand(PyinfraCommand):
    def __init__(self, src: str, dest: str, remote_temp_filename=None, **kwargs):
        super(FileDownloadCommand, self).__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.remote_temp_filename = remote_temp_filename

    def __repr__(self):
        return "FileDownloadCommand({0}, {1})".format(self.src, self.dest)

    def execute(self, state: "State", host: "Host", executor_kwargs):
        executor_kwargs.update(self.executor_kwargs)

        return host.get_file(
            self.src,
            self.dest,
            remote_temp_filename=self.remote_temp_filename,
            print_output=state.print_output,
            print_input=state.print_input,
            **executor_kwargs,
        )


class FunctionCommand(PyinfraCommand):
    def __init__(self, function, args, func_kwargs, **kwargs):
        super(FunctionCommand, self).__init__(**kwargs)
        self.function = function
        self.args = args
        self.kwargs = func_kwargs

    def __repr__(self):
        return "FunctionCommand({0}, {1}, {2})".format(
            self.function.__name__,
            self.args,
            self.kwargs,
        )

    def execute(self, state: "State", host: "Host", executor_kwargs):
        argspec = getfullargspec(self.function)
        if "state" in argspec.args and "host" in argspec.args:
            return self.function(state, host, *self.args, **self.kwargs)

        def execute_function():
            with ctx_config.use(state.config.copy()):
                with ctx_host.use(host):
                    self.function(*self.args, **self.kwargs)

        greenlet = gevent.spawn(execute_function)
        return greenlet.get()


class RsyncCommand(PyinfraCommand):
    def __init__(self, src: str, dest: str, flags, **kwargs):
        super(RsyncCommand, self).__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.flags = flags

    def __repr__(self):
        return "RsyncCommand({0}, {1}, {2})".format(self.src, self.dest, self.flags)

    def execute(self, state: "State", host: "Host", executor_kwargs):
        return host.rsync(
            self.src,
            self.dest,
            self.flags,
            print_output=state.print_output,
            print_input=state.print_input,
            **executor_kwargs,
        )



"""
The files facts provide information about the filesystem and it's contents on the target host.
"""

import re
from datetime import datetime

from pyinfra.api.command import QuoteString, make_formatted_string_command
from pyinfra.api.facts import FactBase
from pyinfra.api.util import try_int

LINUX_STAT_COMMAND = "stat -c 'user=%U group=%G mode=%A atime=%X mtime=%Y ctime=%Z size=%s %N'"
BSD_STAT_COMMAND = "stat -f 'user=%Su group=%Sg mode=%Sp atime=%a mtime=%m ctime=%c size=%z %N%SY'"

STAT_REGEX = (
    r"user=(.*) group=(.*) mode=(.*) "
    r"atime=([0-9]*) mtime=([0-9]*) ctime=([0-9]*) "
    r"size=([0-9]*) (.*)"
)

FLAG_TO_TYPE = {
    "b": "block",
    "c": "character",
    "d": "directory",
    "l": "link",
    "s": "socket",
    "p": "fifo",
    "-": "file",
}

SYMBOL_TO_OCTAL_PERMISSIONS = {
    "rwx": "7",
    "rw-": "6",
    "r-x": "5",
    "r--": "4",
    "-wx": "3",
    "-w-": "2",
    "--x": "1",
}


def _parse_mode(mode):
    """
    Converts ls mode output (rwxrwxrwx) -> integer (755).
    """

    result = ""
    # owner, group, world
    for group in [mode[0:3], mode[3:6], mode[6:9]]:
        if group in SYMBOL_TO_OCTAL_PERMISSIONS:
            result = "{0}{1}".format(result, SYMBOL_TO_OCTAL_PERMISSIONS[group])
        else:
            result = "{0}0".format(result)

    return int(result)


class File(FactBase):
    """
    Returns information about a file on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "mode": 644,
            "size": 3928,
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a file:
        returns ``False``
    """

    type = "file"

    def command(self, path):
        return make_formatted_string_command(
            (
                # only stat if the path exists (file or symlink)
                "! (test -e {0} || test -L {0} ) || "
                "( {linux_stat_command} {0} 2> /dev/null || {bsd_stat_command} {0} )"
            ),
            QuoteString(path),
            linux_stat_command=LINUX_STAT_COMMAND,
            bsd_stat_command=BSD_STAT_COMMAND,
        )

    def process(self, output):
        match = re.match(STAT_REGEX, output[0])
        if not match:
            return None

        data = {}
        path_type = None

        for key, value in (
            ("user", match.group(1)),
            ("group", match.group(2)),
            ("mode", match.group(3)),
            ("atime", match.group(4)),
            ("mtime", match.group(5)),
            ("ctime", match.group(6)),
            ("size", match.group(7)),
        ):
            if key == "mode":
                path_type = FLAG_TO_TYPE[value[0]]
                value = _parse_mode(value[1:])

            elif key == "size":
                value = try_int(value)

            elif key in ("atime", "mtime", "ctime"):
                value = try_int(value)
                if isinstance(value, int):
                    value = datetime.utcfromtimestamp(value)

            data[key] = value

        if path_type != self.type:
            return False

        if path_type == "link":
            filename = match.group(8)
            filename, target = filename.split(" -> ")
            data["link_target"] = target.strip("'").lstrip("`")

        return data


class Link(File):
    """
    Returns information about a link on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "link_target": "/path/to/link/target"
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a link:
        returns ``False``
    """

    type = "link"


class Directory(File):
    """
    Returns information about a directory on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "mode": 644,
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a directory:
        returns ``False``
    """

    type = "directory"


class Socket(File):
    """
    Returns information about a socket on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a socket:
        returns ``False``
    """

    type = "socket"


class Sha1File(FactBase):
    """
    Returns a SHA1 hash of a file. Works with both sha1sum and sha1. Returns
    ``None`` if the file doest not exist.
    """

    _regexes = [
        r"^([a-zA-Z0-9]{40})\s+%s$",
        r"^SHA1\s+\(%s\)\s+=\s+([a-zA-Z0-9]{40})$",
    ]

    def command(self, path):
        self.path = path
        return make_formatted_string_command(
            (
                "test -e {0} && ( "
                "sha1sum {0} 2> /dev/null || shasum {0} 2> /dev/null || sha1 {0} "
                ") || true"
            ),
            QuoteString(path),
        )

    def process(self, output):
        for regex in self._regexes:
            regex = regex % re.escape(self.path)
            matches = re.match(regex, output[0])
            if matches:
                return matches.group(1)


class Sha256File(FactBase):
    """
    Returns a SHA256 hash of a file, or ``None`` if the file does not exist.
    """

    _regexes = [
        r"^([a-zA-Z0-9]{64})\s+%s$",
        r"^SHA256\s+\(%s\)\s+=\s+([a-zA-Z0-9]{64})$",
    ]

    def command(self, path):
        self.path = path
        return make_formatted_string_command(
            (
                "test -e {0} && ( "
                "sha256sum {0} 2> /dev/null "
                "|| shasum -a 256 {0} 2> /dev/null "
                "|| sha256 {0} "
                ") || true"
            ),
            QuoteString(path),
        )

    def process(self, output):
        for regex in self._regexes:
            regex = regex % re.escape(self.path)
            matches = re.match(regex, output[0])
            if matches:
                return matches.group(1)


class Md5File(FactBase):
    """
    Returns an MD5 hash of a file, or ``None`` if the file does not exist.
    """

    _regexes = [
        r"^([a-zA-Z0-9]{32})\s+%s$",
        r"^SHA256\s+\(%s\)\s+=\s+([a-zA-Z0-9]{32})$",
    ]

    def command(self, path):
        self.path = path
        return make_formatted_string_command(
            "test -e {0} && ( md5sum {0} 2> /dev/null || md5 {0} ) || true",
            QuoteString(path),
        )

    def process(self, output):
        for regex in self._regexes:
            regex = regex % re.escape(self.path)
            matches = re.match(regex, output[0])
            if matches:
                return matches.group(1)


class FindInFile(FactBase):
    """
    Checks for the existence of text in a file using grep. Returns a list of matching
    lines if the file exists, and ``None`` if the file does not.
    """

    def command(self, path, pattern, interpolate_variables=False):
        self.exists_flag = "__pyinfra_exists_{0}".format(path)

        if interpolate_variables:
            pattern = '"{0}"'.format(pattern.replace('"', '\\"'))
        else:
            pattern = QuoteString(pattern)

        return make_formatted_string_command(
            (
                "grep -e {0} {1} 2> /dev/null || "
                "( find {1} -type f > /dev/null && echo {2} || true )"
            ),
            pattern,
            QuoteString(path),
            QuoteString(self.exists_flag),
        )

    def process(self, output):
        # If output is the special string: no matches, so return an empty list;
        # this allows us to differentiate between no matches in an existing file
        # or a file not existing.
        if output and output[0] == self.exists_flag:
            return []

        return output


class FindFilesBase(FactBase):
    abstract = True
    default = list

    @staticmethod
    def process(output):
        return output

    def command(self, path, quote_path=True):
        return make_formatted_string_command(
            "find {0} -type {type_flag} || true",
            QuoteString(path) if quote_path else path,
            type_flag=self.type_flag,
        )


class FindFiles(FindFilesBase):
    """
    Returns a list of files from a start path, recursively using ``find``.
    """

    type_flag = "f"


class FindLinks(FindFilesBase):
    """
    Returns a list of links from a start path, recursively using ``find``.
    """

    type_flag = "l"


class FindDirectories(FindFilesBase):
    """
    Returns a list of directories from a start path, recursively using ``find``.
    """

    type_flag = "d"


class Flags(FactBase):
    """
    Returns a list of the file flags set for the specified file or directory.
    """

    requires_command = "chflags"  # don't try to retrieve them if we can't set them

    def command(self, path):
        return make_formatted_string_command(
            "! test -e {0} || stat -f %Sf {0}",
            QuoteString(path),
        )

    def process(self, output):

        return [flag for flag in output[0].split(",") if len(flag) > 0] if len(output) == 1 else []


MARKER_DEFAULT = "# {mark} PYINFRA BLOCK"
MARKER_BEGIN_DEFAULT = "BEGIN"
MARKER_END_DEFAULT = "END"
EXISTS = "__pyinfra_exists_"
MISSING = "__pyinfra_missing_"


class Block(FactBase):
    """
    Returns a (possibly empty) list of the lines found between the markers.

    .. code:: python

        [
            "xray: one",
            "alpha: two"
        ]

    If the ``path`` doesn't exist
        returns ``None``

    If the ``path`` exists but the markers are not found
        returns ``[]``
    """

    # if markers aren't found, awk will return 0 and produce no output but we need to
    # distinguish between "markers not found" and "markers found but nothing between them"
    # for the former we use the empty list (created the call to default) and for the latter
    # the list with a single empty string.
    default = list

    def command(self, path, marker=None, begin=None, end=None):

        self.path = path
        start = (marker or MARKER_DEFAULT).format(mark=begin or MARKER_BEGIN_DEFAULT)
        end = (marker or MARKER_DEFAULT).format(mark=end or MARKER_END_DEFAULT)
        if start == end:
            raise ValueError(f"delimiters for block must be different but found only '{start}'")

        backstop = make_formatted_string_command(
            "(find {0} -type f > /dev/null && echo {1} || echo {2} )",
            QuoteString(path),
            QuoteString(f"{EXISTS}{path}"),
            QuoteString(f"{MISSING}{path}"),
        )
        # m_f_s_c inserts blanks in unfortunate places, e.g. after first slash
        cmd = make_formatted_string_command(
            f"awk \\'/{end}/{{{{f=0}}}} f; /{start}/{{{{f=1}}}}\\' {{0}} || {backstop}",
            QuoteString(path),
        )
        return cmd

    def process(self, output):
        if output and (output[0] == f"{EXISTS}{self.path}"):
            return []
        elif output and (output[0] == f"{MISSING}{self.path}"):
            return None
        else:
            return output
